import numpy as np


def test_mrs_charm(mceq_sib21):
    from MCEq.charm_models import MRS_charm

    e_grid = mceq_sib21.e_grid
    cms = mceq_sib21._int_cs
    model = MRS_charm(e_grid, cms)

    # Test sigma_cc
    sigma = model.sigma_cc(e_grid)
    assert sigma.shape == e_grid.shape

    # Test dsig_dx
    x = np.linspace(0.05, 0.6, 10)
    E = 1e5
    dx_vec = model.dsig_dx(x, E)
    dx_scalar = model.dsig_dx(0.1, E)
    assert dx_vec.shape == x.shape
    assert dx_scalar.shape == ()
    assert np.all(dx_vec >= 0)

    # Test D_dist and LambdaC_dist
    D = model.D_dist(x, E, 421)
    L = model.LambdaC_dist(x, E)
    assert D.shape == x.shape
    assert L.shape == x.shape
    assert np.all(D >= 0)
    assert np.all(L >= 0)

    # Check yield matrix for valid pair
    mat = model.get_yield_matrix(2212, 421)
    assert mat.shape == (len(e_grid), len(e_grid))
    assert np.any(mat > 0)

    # Check yield matrix for invalid projectile
    zero_mat = model.get_yield_matrix(11, 421)
    assert np.allclose(zero_mat, 0)

    # Check yield matrix for invalid secondary
    zero_mat2 = model.get_yield_matrix(2212, 13)
    assert np.allclose(zero_mat2, 0)

    # Check sign condition for LambdaC
    zero_mat3 = model.get_yield_matrix(2212, -4122)
    assert np.allclose(zero_mat3, 0)

    # Test all combinations
    for proj in model.allowed_proj:
        for sec in model.allowed_sec:
            _ = model.get_yield_matrix(proj, sec)



import os
import sysconfig
from ctypes import c_double, cdll

base = os.path.dirname(os.path.abspath(__file__))
suffix = sysconfig.get_config_var("EXT_SUFFIX")
# Some Python 2.7 versions don't define EXT_SUFFIX
if suffix is None and "SO" in sysconfig.get_config_vars():
    suffix = sysconfig.get_config_var("SO")

assert suffix is not None, "Shared lib suffix was not identified."

for fn in os.listdir(base):
    if "libcorsikaatm" in fn and fn.endswith(suffix):
        corsika_acc = cdll.LoadLibrary(os.path.join(base, fn))
        break

for func in [
    corsika_acc.corsika_get_density,
    corsika_acc.planar_rho_inv,
    corsika_acc.corsika_get_m_overburden,
]:
    func.restype = c_double


def corsika_get_density(h_cm, a, b, c, t, hl):
    """Wrap arguments for ctypes function"""
    return corsika_acc.corsika_get_density(
        c_double(h_cm), a.ctypes, b.ctypes, c.ctypes, t.ctypes, hl.ctypes
    )


def planar_rho_inv(X, cos_theta, a, b, c, t, hl):
    """Wrap arguments for ctypes function"""
    return corsika_acc.planar_rho_inv(
        c_double(X),
        c_double(cos_theta),
        a.ctypes,
        b.ctypes,
        c.ctypes,
        t.ctypes,
        hl.ctypes,
    )


def corsika_get_m_overburden(h_cm, a, b, c, t, hl):
    """Wrap arguments for ctypes function"""
    return corsika_acc.corsika_get_m_overburden(
        c_double(h_cm), a.ctypes, b.ctypes, c.ctypes, t.ctypes, hl.ctypes
    )



import copy

import numpy as np
import pytest

from MCEq import config
from MCEq.solvers import solv_numpy


def test_solv_numpy_runs(toy_solver_problem):
    phi0 = toy_solver_problem[-2].copy()
    grid_idcs = toy_solver_problem[-1]

    solution, grid_sol = solv_numpy(*toy_solver_problem)
    assert solution.shape == phi0.shape
    assert grid_sol.shape == (len(grid_idcs), phi0.shape[0])
    assert not np.isnan(solution).any()

    for i in range(1, grid_sol.shape[0]):
        assert np.all(grid_sol[i] <= grid_sol[i - 1])


@pytest.mark.skipif(not config.has_cuda, reason="CUDA not available")
def test_solv_CUDA_sparse_matches_numpy(toy_solver_problem):
    from MCEq.solvers import CUDASparseContext, solv_CUDA_sparse

    solution_numpy, _ = solv_numpy(*toy_solver_problem)

    int_m = toy_solver_problem[3]
    dec_m = toy_solver_problem[4]

    ctx = CUDASparseContext(int_m, dec_m, device_id=config.cuda_gpu_id)

    solution_cuda, _ = solv_CUDA_sparse(
        toy_solver_problem[0],  # nsteps
        toy_solver_problem[1],  # dX
        toy_solver_problem[2],  # rho_inv
        ctx,  # CUDASparseContext
        toy_solver_problem[5],  # phi
        toy_solver_problem[6],  # grid_idcs
    )
    assert solution_cuda == pytest.approx(solution_numpy, rel=1e-5, abs=1e-10)


@pytest.mark.skipif(not config.has_mkl, reason="MKL not available")
def test_solv_MKL_sparse_matches_numpy(toy_solver_problem):
    from MCEq.solvers import solv_MKL_sparse

    toy_solver_problem_mkl = tuple(copy.deepcopy(x) for x in toy_solver_problem)

    solution_numpy, _ = solv_numpy(*toy_solver_problem)

    solution_mkl, _ = solv_MKL_sparse(*toy_solver_problem_mkl)

    assert solution_mkl == pytest.approx(solution_numpy, rel=1e-5, abs=1e-10)


def test_solv_numpy_does_not_modify_input_phi(toy_solver_setup):
    """
    Regression test: ensure solv_numpy doesn't modify the input phi array.

    This was a bug where the NumPy solver modified the input array in place,
    causing subsequent solver calls to start with wrong initial conditions.
    """
    from MCEq.solvers import solv_numpy

    phi_original = toy_solver_setup[-2]
    phi_copy = phi_original.copy()
    # Run solver
    solution, _ = solv_numpy(*toy_solver_setup)

    # The input array should not be modified
    assert np.array_equal(phi_original, phi_copy), (
        "solv_numpy modified the input phi array - this breaks subsequent solver calls"
    )

    # The solution should be different from the input
    assert not np.array_equal(solution, phi_copy), (
        "Solver should produce a different result"
    )


@pytest.mark.skipif(not config.has_cuda, reason="CUDA not available")
def test_cuda_numpy_solver_consistency(toy_solver_setup):
    """
    Regression test: ensure CUDA and NumPy solvers produce consistent results.

    This validates that both solvers implement the same algorithm correctly.

    Note: Requires CuPy >= 12.0.0 for modern sparse matrix interface compatibility.
    """
    from MCEq import config
    from MCEq.solvers import CUDASparseContext, solv_CUDA_sparse

    # Run NumPy solver
    solution_numpy, _ = solv_numpy(*toy_solver_setup)

    # Run CUDA solver
    nsteps, dX, rho_inv, int_m, dec_m, phi, grid_idcs = toy_solver_setup

    ctx = CUDASparseContext(int_m, dec_m, device_id=config.cuda_gpu_id)
    solution_cuda, _ = solv_CUDA_sparse(nsteps, dX, rho_inv, ctx, phi, grid_idcs)

    # Results should match within floating-point precision
    # (CUDA uses float32, NumPy uses float64, so we expect small differences)
    assert np.allclose(
        solution_cuda,
        solution_numpy,
        rtol=1e-5,
        atol=1e-8,
    ), "CUDA and NumPy solvers produce different results"



import atexit
import os
import sysconfig
from ctypes import (
    POINTER,
    c_double,
    c_int,
    c_longlong,
    cdll,
)

base = os.path.dirname(os.path.abspath(__file__))
suffix = sysconfig.get_config_var("EXT_SUFFIX")
# Some Python 2.7 versions don't define EXT_SUFFIX
if suffix is None and "SO" in sysconfig.get_config_vars():
    suffix = sysconfig.get_config_var("SO")

assert suffix is not None, "Shared lib suffix was not identified."

for fn in os.listdir(base):
    if "libspacc" in fn and fn.endswith(suffix):
        spacc = cdll.LoadLibrary(os.path.join(base, fn))
        break

# Declaration of function argument types
spacc.test.restype = c_int

spacc.free_mstore.restype = None

spacc.free_mstore_at.restype = None
spacc.free_mstore_at.argtypes = [c_int]

spacc.create_sparse_matrix.restype = c_int
spacc.create_sparse_matrix.argtypes = [
    c_int,
    c_int,
    c_int,
    c_int,
    POINTER(c_longlong),
    POINTER(c_longlong),
    POINTER(c_double),
]

spacc.gemv.restype = c_int
spacc.gemv.argtypes = [
    c_double,
    c_int,
    POINTER(c_double),
    POINTER(c_double),
]

spacc.daxpy.restype = None
spacc.daxpy.argtypes = [
    c_int,
    c_double,
    POINTER(c_double),
    POINTER(c_double),
]

daxpy = spacc.daxpy

# Initialize
spacc.free_mstore()
# On module unload or reload, free the pointers
atexit.register(spacc.free_mstore)


class SpaccMatrix(object):
    def __init__(self, scipy_sparse_matrix):
        spm = scipy_sparse_matrix.tocoo()
        self.dim_rows, self.dim_cols = spm.shape
        self.nnz = spm.nnz
        self.col = spm.col.astype("longlong")
        self.row = spm.row.astype("longlong")
        self.data = spm.data.astype("double")
        self.store_id = None
        self._create_matrix()

    def __del__(self):
        spacc.free_mstore_at(self.store_id)

    def _create_matrix(self):
        self.store_id = spacc.create_sparse_matrix(
            -1,
            self.dim_rows,
            self.dim_cols,
            self.nnz,
            self.row.ctypes.data_as(POINTER(c_longlong)),
            self.col.ctypes.data_as(POINTER(c_longlong)),
            self.data.ctypes.data_as(POINTER(c_double)),
        )
        if self.store_id < 0:
            raise Exception("Matrix creation failed.")

    def gemv_ctargs(self, alpha, cx, cy):
        """General Matrix-Vector multiplication, expects arguments
        in correct ctypes.

        Performs y = alpha*M*x + y.
        No dimensional checks are performed.
        """
        if spacc.gemv(alpha, self.store_id, cx, cy) != 0:
            raise Exception("Sparse matrix-vector multiplication failed.")

    def gemv_npargs(self, alpha, x, y):
        """General Matrix-Vector multiplication, expects numpy arrays as arguments.

        Performs y = alpha*M*x + y.
        """
        assert x.shape[0] == self.dim_cols
        assert y.shape[0] == self.dim_cols
        assert x.dtype == "float64"
        assert y.dtype == "float64"
        alpha = float(alpha)

        cy = y.ctypes.data_as(POINTER(c_double))
        cx = x.ctypes.data_as(POINTER(c_double))
        if spacc.gemv(alpha, self.store_id, cx, cy) != 0:
            raise Exception("Sparse matrix-vector multiplication failed.")



import numpy as np
import pytest

import MCEq.geometry.density_profiles as dp

corsika_expected = [
    ("USStd", None, (1036.099233683902, 0.00015623258808300557)),
    ("BK_USStd", None, (1033.8094962133184, 0.00015782685585891685)),
    ("Karlsruhe", None, (1055.861981113731, 0.00016209949387937668)),
    ("ANTARES/KM3NeT-ORCA", "Summer", (986.9593811082788, 0.00015529574727367941)),
    ("ANTARES/KM3NeT-ORCA", "Winter", (988.4293864278521, 0.0001589317236294479)),
    ("KM3NeT-ARCA", "Summer", (1032.7184058861765, 0.00016954131888323744)),
    ("KM3NeT-ARCA", "Winter", (1039.3697214845179, 0.00016202068935405075)),
    ("KM3NeT", None, (1018.1547240905948, 0.0001609490344992944)),
    ("SouthPole", "December", (1011.4568036341923, 0.00014626903051217024)),
    ("SouthPole", "June", (1020.3505579524912, 0.00018246219074986874)),
    ("PL_SouthPole", "January", (1019.974568696789, 0.0001464549375212421)),
    ("PL_SouthPole", "August", (1019.9764946890782, 0.0001685608228906579)),
    ("SDR_SouthPole", "January", (1034.0143423913353, 0.00014632473385006882)),
    ("SDR_SouthPole", "February", (1035.9936617195242, 0.00014918253835319762)),
    ("SDR_SouthPole", "March", (1038.9308255875999, 0.00015298022770783904)),
    ("SDR_SouthPole", "April", (1041.9777521676683, 0.0001601901839449753)),
    ("SDR_SouthPole", "May", (1039.907580402069, 0.00016807635328451188)),
    ("SDR_SouthPole", "June", (1037.2666634880595, 0.00017542702214644906)),
    ("SDR_SouthPole", "July", (1035.7659887947702, 0.00018145735342421507)),
    ("SDR_SouthPole", "August", (1034.5825095745786, 0.0001865567510599077)),
    ("SDR_SouthPole", "September", (1033.8370290916944, 0.00018184168516617518)),
    ("SDR_SouthPole", "October", (1300.9045332765688, 0.0001729949911605945)),
    ("SDR_SouthPole", "November", (1021.2313399252452, 0.00016452181236376558)),
    ("SDR_SouthPole", "December", (1028.3233680860317, 0.00015434578972428948)),
]


# Test that all corsika atmospheres are tested
def test_cka_atm_completeness():
    from MCEq.geometry.atmosphere_parameters import list_available_corsika_atmospheres

    missing = []
    expected_entries = {(loc, season) for loc, season, _ in corsika_expected}
    for loc, season in list_available_corsika_atmospheres():
        if (loc, season) not in expected_entries:
            missing.append((loc, season))
    if missing:
        for i, (loc, season) in enumerate(missing):
            # Create reference data
            from MCEq.geometry.density_profiles import CorsikaAtmosphere

            cka_obj = CorsikaAtmosphere(loc, season)
            ref = (float(cka_obj.max_X), float(1.0 / cka_obj.r_X2rho(100.0)))
            missing[i] = (loc, season, ref)

    assert len(missing) == 0, f"Missing tests for {missing}."


ids = [f"{loc}-{season or 'None'}" for loc, season, _ in corsika_expected]


@pytest.mark.parametrize(("loc", "season", "expected"), corsika_expected, ids=ids)
def test_corsika_atm(loc, season, expected):
    from MCEq.geometry.density_profiles import CorsikaAtmosphere

    cka_obj = CorsikaAtmosphere(loc, season)
    assert np.allclose([cka_obj.max_X, 1.0 / cka_obj.r_X2rho(100.0)], expected)


msis00_expected = [
    ("SouthPole", "January", (1022.6914983678925, 0.00014380042112573175)),
    ("Karlsruhe", "January", (1041.2180457811605, 0.00016046129606232836)),
    ("Geneva", "January", (1044.6608866969684, 0.00016063221634835724)),
    ("Tokyo", "January", (1046.427667371285, 0.00016041531186210874)),
    ("SanGrasso", "January", (1048.6505423154006, 0.00016107650347480857)),
    ("TelAviv", "January", (1050.6431802896034, 0.00016342084740033518)),
    ("KSC", "January", (1050.2145039327452, 0.00016375664772178006)),
    ("SoudanMine", "January", (1033.3640270683418, 0.00015614485659072835)),
    ("Tsukuba", "January", (1045.785578319159, 0.00015970449150213374)),
    ("LynnLake", "January", (1019.9475650272982, 0.000153212909250962)),
    ("PeaceRiver", "January", (1020.3640351872195, 0.00015221038616604717)),
    ("FtSumner", "January", (1047.964376368261, 0.00016218804771381842)),
    ("SouthPole", "July", (1022.1737895082897, 0.00017812023753792838)),
]

ids = [f"{loc}-{season or 'None'}" for loc, season, _ in msis00_expected]


@pytest.mark.parametrize(("loc", "season", "expected"), msis00_expected, ids=ids)
def test_msis_atm(loc, season, expected):
    from MCEq.geometry.density_profiles import MSIS00Atmosphere

    msis_obj = MSIS00Atmosphere(loc, season)
    if expected is None:
        ref = (float(msis_obj.max_X), float(1.0 / msis_obj.r_X2rho(100.0)))
        msg = f"MSIS-00 reference data for {loc} in {season} not available."
        msg += f" Creating a new one. {ref}"
        pytest.fail(msg)
    assert np.allclose([msis_obj.max_X, 1.0 / msis_obj.r_X2rho(100.0)], expected)


@pytest.mark.parametrize(
    "cls,args",
    [
        (dp.CorsikaAtmosphere, ("USStd", None)),
        (dp.MSIS00Atmosphere, ("SouthPole", "January")),
        (dp.IsothermalAtmosphere, ("Nowhere", None)),
    ],
)
def test_common_atmosphere_interface(cls, args):
    atm = cls(*args)
    atm.set_theta(0.0)

    X_test = np.linspace(1, atm.max_X * 0.99, 10)
    h_test = np.linspace(0, atm.geom.h_atm, 10)

    # r_X2rho should give positive finite values
    inv_rho = atm.r_X2rho(X_test)
    assert np.all(np.isfinite(inv_rho))
    assert np.all(inv_rho > 0)

    # X2rho should be consistent with inverse
    rho = atm.X2rho(X_test)
    assert np.allclose(1 / rho, inv_rho, rtol=1e-3)

    # h2X and X2h roundtrip
    for h in h_test:
        X = atm.h2X(h)
        h_back = atm.X2h(X)
        assert np.isclose(h, h_back, rtol=1e-2)

    # misc functions
    for h in h_test:
        mol = atm.moliere_air(h)
        nrel = atm.nref_rel_air(h)
        theta_c = atm.theta_cherenkov_air(h)
        gamma_c = atm.gamma_cherenkov_air(h)
        assert mol > 0
        assert 0 < nrel < 1e-3
        assert 0 < theta_c < 90
        assert gamma_c > 1


def test_corsika_depth_and_inverse_functions():
    atm = dp.CorsikaAtmosphere("USStd", None)
    atm.set_theta(0.0)

    # depth2height should roughly invert get_mass_overburden
    h_cm = np.linspace(1e5, 6e5, 10)
    X = [atm.get_mass_overburden(h) for h in h_cm]
    h_back = [atm.depth2height(x) for x in X]
    assert np.allclose(h_cm, h_back, rtol=0.1)

    # rho_inv gives inverse density
    rho = [atm.get_density(h) for h in h_cm]
    rho_inv = [atm.rho_inv(atm.get_mass_overburden(h), np.cos(0.0)) for h in h_cm]
    assert np.allclose([1.0 / r for r in rho], rho_inv, rtol=1e-2)

    # calc_thickl returns 5 values
    thickl = atm.calc_thickl()
    assert isinstance(thickl, list)
    assert len(thickl) == 5


def test_isothermal_mass_overburden():
    atm = dp.IsothermalAtmosphere("Unknown", None)
    h = np.linspace(0, 5e5, 10)
    overburden = atm.get_mass_overburden(h)
    assert np.all(overburden > 0)
    assert np.all(overburden <= atm.X0)


def test_msis_setters_and_cache_clear():
    atm = dp.MSIS00Atmosphere("SouthPole", "January")
    atm.theta_deg = 42
    atm._clear_cache()
    assert atm.theta_deg is None

    atm.set_location("Karlsruhe")
    atm.set_location_coord(10.0, 48.0)
    atm.set_season("February")
    atm.set_doy(42)
    assert atm.theta_deg is None  # should still be cleared

    h = 2e5
    T = atm.get_temperature(h)
    assert 100 < T < 1000


def test_msis00_icecube_centered():
    atm = dp.MSIS00IceCubeCentered(
        "Karlsruhe", "January"
    )  # should override to SouthPole
    # just check if works
    assert atm.get_density(1e5) > 0

    # test latitude at 0 and 90 deg
    lat_0 = atm._latitude(0.0)
    lat_90 = atm._latitude(90.0)
    assert -90.0 <= lat_0 <= 0.0
    assert -90.0 <= lat_90 <= 0.0
    assert lat_0 < lat_90  # as zenith increases, impact moves away from vertical

    # test set_theta for downgoing and upgoing
    atm.set_theta(45.0)
    assert atm.theta_deg == 45.0

    atm.set_theta(135.0)
    assert atm.theta_deg == 135.0


@pytest.mark.xfail(reason="AIRSAtmosphere requires unavailable data files")
def test_airs_instantiation():
    dp.AIRSAtmosphere("SouthPole", "January")


@pytest.mark.parametrize("X", [1.0, 10.0, 100.0])
def test_generalized_target(X):
    tgt = dp.GeneralizedTarget(len_target=100.0, env_density=2.0, env_name="default")

    # reset + basic props
    assert tgt.max_den == 2.0
    assert np.isclose(tgt.max_X, 100.0 * 2.0)
    assert np.isclose(tgt.get_density_X(X), 2.0)
    assert np.isclose(tgt.r_X2rho(X), 0.5)
    assert np.isclose(tgt.get_density(tgt.s_X2h(X)), 2.0)

    # set_length ok
    tgt.set_length(150.0)
    assert tgt.len_target == 150.0
    assert np.isclose(tgt.max_X, 2.0 * 150.0)

    # add material (replace default)
    tgt.reset()
    tgt.add_material(0.0, 3.0, "iron")
    assert len(tgt.mat_list) == 1
    assert np.isclose(tgt.get_density(0), 3.0)

    # add second layer
    tgt.reset()
    tgt.add_material(60.0, 4.0, "lead")
    assert len(tgt.mat_list) == 2
    assert np.isclose(tgt.get_density(70), 4.0)

    # set_length too small
    tgt.reset()
    tgt.add_material(50.0, 3.0, "x")
    with pytest.raises(Exception):
        tgt.set_length(40.0)

    # add invalid materials
    with pytest.raises(Exception):
        tgt.add_material(200.0, 1.0, "fail")
    with pytest.raises(Exception):
        tgt.add_material(30.0, 2.0, "bad")  # not monotonic

    # set_theta must raise
    with pytest.raises(NotImplementedError):
        tgt.set_theta(0.0)

    # spline roundtrip
    X_vals = np.linspace(1, tgt.max_X - 1, 10)
    h_vals = tgt.s_X2h(X_vals)
    assert np.allclose(tgt.s_h2X(h_vals), X_vals)

    # density errors
    with pytest.raises(Exception):
        tgt.get_density([0, 200])
    with pytest.raises(Exception):
        tgt.get_density_X(tgt.max_X * 1.01)

    # reset again
    tgt.add_material(60.0, 5.0, "z")
    tgt.reset()
    assert len(tgt.mat_list) == 1
    assert np.isclose(tgt.get_density(10), 2.0)



import numpy as np

from MCEq import config


class EarthGeometry:
    r"""A model of the Earth's geometry, approximating it
       by a sphere. The figure below illustrates the meaning of the parameters.

    .. figure:: ../_static/graphics/geometry.*
        :scale: 50 %
        :alt: picture of geometry


        Curved geometry as it is used in the code (not to scale!).

    Example:
      The plots below will be produced by executing the module::

          $ python geometry.py

    .. plot::

        import matplotlib.pyplot as plt
        import numpy as np
        from MCEq.geometry.geometry import *

        earth = EarthGeometry()

        theta_list = np.linspace(0, 90, 500)
        h_vec = np.linspace(0, earth.h_atm, 500)
        th_list_rad = np.deg2rad(theta_list)
        fig = plt.figure(figsize=(5, 4))
        fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
        plt.plot(theta_list, earth.path_len(th_list_rad) / 1e5, lw=2)
        plt.xlabel(r"zenith $\theta$ at detector")
        plt.ylabel(r"path length $l(\theta)$ in km")
        ax = plt.gca()
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.xaxis.set_ticks_position("bottom")
        ax.yaxis.set_ticks_position("left")

        fig = plt.figure(figsize=(5, 4))
        fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
        plt.plot(
            theta_list, np.arccos(earth.cos_th_star(th_list_rad)) / np.pi * 180.0, lw=2
        )
        plt.xlabel(r"zenith $\theta$ at detector")
        plt.ylabel(r"$\theta^*$ at top of the atm.")
        plt.ylim([0, 90])
        ax = plt.gca()
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.xaxis.set_ticks_position("bottom")
        ax.yaxis.set_ticks_position("left")

        fig = plt.figure(figsize=(5, 4))
        fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
        plt.plot(h_vec / 1e5, earth.delta_l(h_vec, np.deg2rad(85.0)) / 1e5, lw=2)
        plt.ylabel(r"Path length $\Delta l(h)$ in km")
        plt.xlabel(r"atm. height $h_{atm}$ in km")
        ax = plt.gca()
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.xaxis.set_ticks_position("bottom")
        ax.yaxis.set_ticks_position("left")

        fig = plt.figure(figsize=(5, 4))
        fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
        for theta in [30.0, 60.0, 70.0, 80.0, 85.0, 90.0]:
            theta_path = np.deg2rad(theta)
            delta_l_vec = np.linspace(0, earth.path_len(theta_path), 1000)
            plt.plot(
                delta_l_vec / 1e5,
                earth.h(delta_l_vec, theta_path) / 1e5,
                label=rf"${theta}^o$",
                lw=2,
            )
        plt.legend()
        plt.xlabel(r"path length $\Delta l$ [km]")
        plt.ylabel(r"atm. height $h_{atm}(\Delta l, \theta)$ [km]")
        ax = plt.gca()
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.xaxis.set_ticks_position("bottom")
        ax.yaxis.set_ticks_position("left")
        plt.show()

    Attributes:
      h_obs (float): observation level height [cm]
      h_atm (float): top of the atmosphere [cm]
      r_E (float): radius Earth [cm]
      r_top (float): radius at top of the atmosphere  [cm]
      r_obs (float): radius at observation level [cm]

    """

    def __init__(self):
        self.h_obs = config.h_obs * 1e2  # cm
        self.h_atm = config.h_atm * 1e2  # cm
        self.r_E = config.r_E * 1e2  # cm
        self.r_top = self.r_E + self.h_atm
        self.r_obs = self.r_E + self.h_obs

    def _A_1(self, theta):
        r"""Segment length :math:`A1(\theta)` in cm."""
        return self.r_obs * np.cos(theta)

    def _A_2(self, theta):
        r"""Segment length :math:`A2(\theta)` in cm."""
        return self.r_obs * np.sin(theta)

    def path_len(self, theta):
        r"""Returns path length in [cm] for given zenith
        angle :math:`\theta` [rad].
        """
        return np.sqrt(self.r_top**2 - self._A_2(theta) ** 2) - self._A_1(theta)

    def cos_th_star(self, theta):
        r"""Returns the zenith angle at atmospheric boarder
        :math:`\cos(\theta^*)` in [rad] as a function of zenith at detector.
        """
        return (self._A_1(theta) + self.path_len(theta)) / self.r_top

    def h(self, dl, theta):
        r"""Height above surface at distance :math:`dl` counted from the beginning
        of path :math:`l(\theta)` in cm.
        """
        return (
            np.sqrt(
                self._A_2(theta) ** 2
                + (self._A_1(theta) + self.path_len(theta) - dl) ** 2
            )
            - self.r_E
        )

    def delta_l(self, h, theta):
        r"""Distance :math:`dl` covered along path :math:`l(\theta)`
        as a function of current height. Inverse to :func:`h`.
        """
        return (
            self._A_1(theta)
            + self.path_len(theta)
            - np.sqrt((h + self.r_E) ** 2 - self._A_2(theta) ** 2)
        )


def chirkin_cos_theta_star(costheta):
    r""":math:`\cos(\theta^*)` parameterization.

    This function returns the equivalent zenith angle for
    for very inclined showers. It is based on a CORSIKA study by
    `D. Chirkin, hep-ph/0407078v1, 2004
    <http://arxiv.org/abs/hep-ph/0407078v1>`_.

    Args:
        costheta (float): :math:`\cos(\theta)` in [rad]

    Returns:
        float: :math:`\cos(\theta*)` in [rad]
    """

    p1 = 0.102573
    p2 = -0.068287
    p3 = 0.958633
    p4 = 0.0407253
    p5 = 0.817285
    x = costheta
    return np.sqrt((x**2 + p1**2 + p2 * x**p3 + p4 * x**p5) / (1 + p1**2 + p2 + p4))


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    earth = EarthGeometry()

    theta_list = np.linspace(0, 90, 500)
    h_vec = np.linspace(0, earth.h_atm, 500)
    th_list_rad = np.deg2rad(theta_list)
    fig = plt.figure(figsize=(5, 4))
    fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
    plt.plot(theta_list, earth.path_len(th_list_rad) / 1e5, lw=2)
    plt.xlabel(r"zenith $\theta$ at detector")
    plt.ylabel(r"path length $l(\theta)$ in km")
    ax = plt.gca()
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")

    fig = plt.figure(figsize=(5, 4))
    fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
    plt.plot(
        theta_list, np.arccos(earth.cos_th_star(th_list_rad)) / np.pi * 180.0, lw=2
    )
    plt.xlabel(r"zenith $\theta$ at detector")
    plt.ylabel(r"$\theta^*$ at top of the atm.")
    plt.ylim([0, 90])
    ax = plt.gca()
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")

    fig = plt.figure(figsize=(5, 4))
    fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
    plt.plot(h_vec / 1e5, earth.delta_l(h_vec, np.deg2rad(85.0)) / 1e5, lw=2)
    plt.ylabel(r"Path length $\Delta l(h)$ in km")
    plt.xlabel(r"atm. height $h_{atm}$ in km")
    ax = plt.gca()
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")

    fig = plt.figure(figsize=(5, 4))
    fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
    for theta in [30.0, 60.0, 70.0, 80.0, 85.0, 90.0]:
        theta_path = np.deg2rad(theta)
        delta_l_vec = np.linspace(0, earth.path_len(theta_path), 1000)
        plt.plot(
            delta_l_vec / 1e5,
            earth.h(delta_l_vec, theta_path) / 1e5,
            label=rf"${theta}^o$",
            lw=2,
        )
    plt.legend()
    plt.xlabel(r"path length $\Delta l$ [km]")
    plt.ylabel(r"atm. height $h_{atm}(\Delta l, \theta)$ [km]")
    ax = plt.gca()
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")
    plt.show()



import sys

import crflux.models as pm
import numpy as np
import pytest
from pytest import approx

import MCEq.geometry.density_profiles as dprof
from MCEq.geometry.atmosphere_parameters import list_available_corsika_atmospheres

if sys.platform.startswith("win") and sys.maxsize <= 2**32:
    pytest.skip("Skip model test on 32-bit Windows.", allow_module_level=True)


def test_solve_default(mceq_sib21):
    mceq_sib21.solve()
    sol = mceq_sib21.get_solution("mu+", mag=0, integrate=True)
    assert sol is not None

    assert np.sum(mceq_sib21._phi0) != np.sum(mceq_sib21._solution)

    muons = mceq_sib21.get_solution("total_mu-", mag=0)

    assert np.sum(muons) != 0


def test_solve_skip_integration_path(mceq_sib21):
    mceq_sib21._calculate_integration_path(int_grid=None, grid_var="X")
    mceq_sib21.solve(skip_integration_path=True)
    sol = mceq_sib21.get_solution("mu+", mag=0, integrate=True)
    assert sol is not None


def test_solve_other_grid_var(mceq_sib21):
    with pytest.raises(NotImplementedError):
        mceq_sib21.solve(grid_var="Y")


@pytest.mark.parametrize(
    ["int_grid", "grid_shape"],
    [[None, (0,)], [[0, 1], (2, 2232)]],
    ids=["no-grid", "with-grid"],
)
def test_solve_int_grid(mceq_sib21, int_grid, grid_shape):
    mceq_sib21.solve(int_grid)
    assert mceq_sib21.grid_sol.shape == grid_shape


@pytest.mark.parametrize(
    ["leading_process", "lenX"],
    [
        ["decays", 1248],
        ["auto", 1248],
        ["interactions", 1248],
    ],
)
def test_integration_path_leading_process(mceq_sib21, leading_process, lenX):
    """Fix this test by resolving issue #66"""
    from MCEq import config

    config.leading_process = leading_process
    int_grid = None
    grid_var = "X"
    mceq_sib21._calculate_integration_path(int_grid, grid_var)
    integration_path = mceq_sib21.integration_path
    assert integration_path[0] == lenX


def test_integration_path_grid_idcs(mceq_sib21):
    int_grid = [0, 1]

    grid_var = "X"
    mceq_sib21._calculate_integration_path(int_grid, grid_var)
    integration_path = mceq_sib21.integration_path
    grid_idcs = integration_path[-1]
    assert len(grid_idcs) == 2
    assert grid_idcs[0] == 0
    assert grid_idcs[0] < grid_idcs[1]


testdata_theta = [
    [0.0, 8.832749837571275e-08],
    [30.0, 9.907914648457978e-08],
    [60.0, 1.5032860734447641e-07],
]

ids_theta = [f"{th[0]}" for th in testdata_theta]


@pytest.mark.parametrize(["theta", "nmu"], testdata_theta, ids=ids_theta)
def test_set_theta_deg(mceq_sib21, theta, nmu):
    mceq_sib21.set_theta_deg(theta)
    mceq_sib21.solve()
    nmu_sol = np.sum(
        mceq_sib21.get_solution("mu+", mag=0, integrate=True)
        + mceq_sib21.get_solution("mu-", mag=0, integrate=True)
    )
    assert nmu_sol == approx(nmu, abs=1e-8)


testdata_model = [
    ["QGSJETII04", 48],
    ["SIBYLL21", 52],
]
ids_model = [f"{model[0]}" for model in testdata_model]


@pytest.mark.parametrize(["model", "n"], testdata_model, ids=ids_model)
def test_set_interaction_model_model(mceq_sib21, model, n):
    mceq_sib21.set_interaction_model(model)
    n_particles = len(mceq_sib21._particle_list)
    assert n_particles == n


def test_set_interaction_model_update_particle_list(mceq_sib21):
    n_particles_sib = len(mceq_sib21._particle_list)

    mceq_sib21.set_interaction_model("QGSJETII04", update_particle_list=True)
    n_particles = len(mceq_sib21._particle_list)
    assert n_particles == 48

    mceq_sib21.set_interaction_model("SIBYLL21", update_particle_list=True)
    n_particles_s = len(mceq_sib21._particle_list)

    assert n_particles_s == n_particles_sib


@pytest.mark.parametrize(
    ["particle_list", "projectiles"],
    [
        [None, 13],
        [[(2212, 0)], 1],
    ],
    ids=["None", "proton"],
)
def test_mceq_init_particles_list(particle_list, projectiles):
    import crflux.models as pm

    from MCEq.core import MCEqRun

    mceq_sib21 = MCEqRun(
        interaction_model="SIBYLL21",
        theta_deg=0.0,
        primary_model=(pm.HillasGaisser2012, "H3a"),
        particle_list=particle_list,
    )

    assert (
        sum([p.is_projectile for p in mceq_sib21.pman.cascade_particles]) == projectiles
    )


testdata_primary = [
    [1e3, 1.188606352544364e-05, 2.3833836497715642e-07, -4.389720437130184e-08],
    [1e4, 0.09924074296757761, 0.024548469717221015, 0.001458070555523793],
    [1e5, 0.9113452102164941, 0.2824779229726991, 0.020207277840290347],
]
ids_primary = [f"energy={primary[0]}" for primary in testdata_primary]


@pytest.mark.parametrize(
    ["energy", "nmu", "nnumu", "nnue"], testdata_primary, ids=ids_primary
)
def test_single_primary(mceq_sib21, energy, nmu, nnumu, nnue):
    mceq_sib21.set_interaction_model("SIBYLL21", force=True)
    mceq_sib21.set_theta_deg(0.0)
    mceq_sib21.set_single_primary_particle(E=energy, pdg_id=2212)
    mceq_sib21.solve()
    nmu_sol = np.sum(
        mceq_sib21.get_solution("mu+", mag=0, integrate=True)
        + mceq_sib21.get_solution("mu-", mag=0, integrate=True)
    )
    nnumu_sol = np.sum(
        mceq_sib21.get_solution("numu", mag=0, integrate=True)
        + mceq_sib21.get_solution("antinumu", mag=0, integrate=True)
    )
    nnue_sol = np.sum(
        mceq_sib21.get_solution("nue", mag=0, integrate=True)
        + mceq_sib21.get_solution("antinue", mag=0, integrate=True)
    )

    assert nmu_sol == approx(nmu, rel=5e-3)
    assert nnumu_sol == approx(nnumu, rel=5e-3)
    assert nnue_sol == approx(nnue, rel=5e-3)


def test_single_primary_pdg_corsika(mceq_sib21):
    mceq_sib21.set_interaction_model("SIBYLL21", force=True)
    mceq_sib21.set_theta_deg(0.0)
    mceq_sib21.set_single_primary_particle(E=1e5, pdg_id=1000020040)
    mceq_sib21.solve()

    pdg_sol = mceq_sib21.get_solution("mu+", mag=0, integrate=True)

    mceq_sib21.set_single_primary_particle(E=1e5, corsika_id=402)
    mceq_sib21.solve()

    corsika_sol = mceq_sib21.get_solution("mu+", mag=0, integrate=True)

    assert np.allclose(pdg_sol, corsika_sol)


def test_single_primary_e_too_low(mceq_sib21):
    mceq_sib21.set_interaction_model("SIBYLL21", force=True)
    mceq_sib21.set_theta_deg(0.0)
    with pytest.raises(Exception):
        mceq_sib21.set_single_primary_particle(E=1e0, pdg_id=2212)


testdata_ecenters = [
    8.91250938e2,
    1.12201845e3,
    1.41253754e3,
    1.77827941e3,
    2.23872114e3,
    2.81838293e3,
    3.54813389e3,
    4.46683592e3,
    5.62341325e3,
    7.07945784e3,
    8.91250938e3,
    1.12201845e4,
    1.41253754e4,
    1.77827941e4,
    2.23872114e4,
    2.81838293e4,
    3.54813389e4,
    4.46683592e4,
    5.62341325e4,
    7.07945784e4,
    8.91250938e4,
    1.12201845e5,
    1.41253754e5,
    1.77827941e5,
    2.23872114e5,
    2.81838293e5,
    3.54813389e5,
    4.46683592e5,
    5.62341325e5,
    7.07945784e5,
    8.91250938e5,
]
testdata_eedges = [
    7.94328235e2,
    1.00000000e3,
    1.25892541e3,
    1.58489319e3,
    1.99526231e3,
    2.51188643e3,
    3.16227766e3,
    3.98107171e3,
    5.01187234e3,
    6.30957344e3,
    7.94328235e3,
    1.00000000e4,
    1.25892541e4,
    1.58489319e4,
    1.99526231e4,
    2.51188643e4,
    3.16227766e4,
    3.98107171e4,
    5.01187234e4,
    6.30957344e4,
    7.94328235e4,
    1.00000000e5,
    1.25892541e5,
    1.58489319e5,
    1.99526231e5,
    2.51188643e5,
    3.16227766e5,
    3.98107171e5,
    5.01187234e5,
    6.30957344e5,
    7.94328235e5,
    1.00000000e6,
]
testdata_ewidths = [
    2.05671765e2,
    2.58925412e2,
    3.25967781e2,
    4.10369123e2,
    5.16624117e2,
    6.50391229e2,
    8.18794045e2,
    1.03080063e3,
    1.29770111e3,
    1.63370890e3,
    2.05671765e3,
    2.58925412e3,
    3.25967781e3,
    4.10369123e3,
    5.16624117e3,
    6.50391229e3,
    8.18794045e3,
    1.03080063e4,
    1.29770111e4,
    1.63370890e4,
    2.05671765e4,
    2.58925412e4,
    3.25967781e4,
    4.10369123e4,
    5.16624117e4,
    6.50391229e4,
    8.18794045e4,
    1.03080063e5,
    1.29770111e5,
    1.63370890e5,
    2.05671765e5,
]


def test_energy_grid_access(mceq_sib21):
    print(mceq_sib21.e_grid)
    print(mceq_sib21.e_bins)
    print(mceq_sib21.e_widths)
    assert np.allclose(mceq_sib21.e_grid, testdata_ecenters, rtol=1e-8, atol=0)
    assert np.allclose(mceq_sib21.e_bins, testdata_eedges, rtol=1e-8, atol=0)
    assert np.allclose(mceq_sib21.e_widths, testdata_ewidths, rtol=1e-8, atol=0)


def test_closest_energy(mceq_sib21):
    closest_energy = np.array(
        [mceq_sib21.closest_energy(energy) for energy in testdata_ecenters]
    )
    assert np.allclose(closest_energy, testdata_ecenters, rtol=0, atol=1e1)


def test_get_solution_grid_idx(mceq_sib21):
    import crflux.models as pm

    mceq_sib21.set_primary_model(pm.HillasGaisser2012, "H3a")
    mceq_sib21.solve()

    with pytest.raises(Exception):
        mceq_sib21.get_solution("mu+", mag=0, integrate=True, grid_idx=0)

    mceq_sib21.solve([0, 1])
    sol0 = mceq_sib21.get_solution("mu+", mag=0, integrate=True, grid_idx=0)
    sol1 = mceq_sib21.get_solution("mu+", mag=0, integrate=True, grid_idx=1)
    sol_last = mceq_sib21.get_solution("mu+", mag=0, integrate=True, grid_idx=100)

    assert np.allclose(sol0, 0)
    assert np.all(sol1 != 0)
    assert np.allclose(sol_last, sol1)


def test_get_solution_wrong_particle_name(mceq_sib21):
    from MCEq import config

    config.excpt_on_missing_particle = True
    mceq_sib21.solve()
    with pytest.raises(Exception):
        mceq_sib21.get_solution("proton", mag=0, integrate=True, grid_idx=0)


def test_get_solution_prefixes(mceq_sib21):
    mceq_sib21.solve()
    total = mceq_sib21.get_solution("total_mu+", mag=0, integrate=True)
    conv = mceq_sib21.get_solution("conv_mu+", mag=0, integrate=True)
    prompt = mceq_sib21.get_solution("pr_mu+", mag=0, integrate=True)

    assert np.allclose(total, conv + prompt)


@pytest.mark.parametrize(
    "return_as",
    ["total energy", "kinetic energy", "total momentum", "none"],
)
@pytest.mark.parametrize("integrate", [True, False])
def test_get_solution_return_as(mceq_sib21, return_as, integrate):
    mceq_sib21.solve()

    if return_as == "none":
        with pytest.raises(Exception):
            result = mceq_sib21.get_solution(
                "mu+", return_as=return_as, integrate=integrate
            )
    else:
        result = mceq_sib21.get_solution(
            "mu+", return_as=return_as, integrate=integrate
        )
        if isinstance(result, tuple):
            xgrid, values = result
            assert len(xgrid) == len(values)
            assert np.all(np.isfinite(values))
        else:
            assert np.all(np.isfinite(result))


def test_get_solution_dont_sum_helicities(mceq_sib21):
    mceq_sib21.solve()

    # Get solution with summed helicities (default)
    solution_summed = mceq_sib21.get_solution("e-", dont_sum_helicities=False)
    solution_left_do = mceq_sib21.get_solution("e-_l", dont_sum_helicities=False)
    solution_right_do = mceq_sib21.get_solution("e-_r", dont_sum_helicities=False)

    # Get individual helicity states
    solution_left = mceq_sib21.get_solution("e-_l", dont_sum_helicities=True)
    solution_right = mceq_sib21.get_solution("e-_r", dont_sum_helicities=True)
    solution_0 = mceq_sib21.get_solution("e-", dont_sum_helicities=True)

    # Manual sum should equal the summed solution
    manual_sum = solution_left + solution_right + solution_0

    assert solution_left == approx(solution_left_do)
    assert solution_right == approx(solution_right_do)
    assert solution_summed == approx(manual_sum)


def test_solve_from_integration_path(mceq_sib21):
    # Normal solve
    mceq_sib21.solve()
    solution_normal = np.copy(mceq_sib21._solution)

    # Get the integration path that was used
    nsteps, dX, rho_inv, grid_idcs = mceq_sib21.integration_path

    # Solve using the same integration path
    mceq_sib21.solve_from_integration_path(nsteps, dX, rho_inv, grid_idcs)
    solution_from_path = mceq_sib21._solution

    # Should give identical results
    assert solution_normal == approx(solution_from_path)


@pytest.mark.parametrize(
    "pdg_id",
    [
        2212,
        2112,
    ],
)
@pytest.mark.parametrize("append", [False, True])
def test_set_initial_spectrum(mceq_sib21, pdg_id, append):
    cond = mceq_sib21._restore_initial_condition
    phi0_backup = mceq_sib21._phi0.copy()

    particle = mceq_sib21.pman[pdg_id]
    phi0 = mceq_sib21._phi0[particle.lidx : particle.uidx].copy()

    spectrum = np.ones_like(phi0) * 10
    mceq_sib21.set_initial_spectrum(spectrum, pdg_id, append)

    phi1 = mceq_sib21._phi0[particle.lidx : particle.uidx]

    if append:
        assert np.allclose(phi0, phi1 - spectrum)
    if not append:
        assert np.allclose(spectrum, phi1)

    mceq_sib21._phi0 = phi0_backup
    mceq_sib21._restore_initial_condition = cond


msis_atmospheres = [
    "SouthPole",
    "Karlsruhe",
    "Geneva",
    "Tokyo",
    "SanGrasso",
    "TelAviv",
    "KSC",
    "SoudanMine",
    "Tsukuba",
    "LynnLake",
    "PeaceRiver",
    "FtSumner",
]

season = ["January", "July"]

corsika_atmospheres = list_available_corsika_atmospheres()


test_densities_cases = []
# MSIS00: all atmospheres
for atmo in msis_atmospheres:
    for s in ["January", "July"]:
        test_densities_cases.append(
            pytest.param("MSIS00", (atmo, s), id=f"MSIS00-{atmo}-{s}")
        )

# MSIS00_IC and AIRS: only SouthPole
for model in ["MSIS00_IC", "AIRS"]:
    for s in ["January", "July"]:
        if model == "AIRS":
            test_densities_cases.append(
                pytest.param(
                    model,
                    ("SouthPole", s),
                    id=f"{model}-SouthPole-{s}",
                    marks=pytest.mark.xfail(reason="Fix issure #71"),
                )
            )
        else:
            test_densities_cases.append(
                pytest.param(model, ("SouthPole", s), id=f"{model}-SouthPole-{s}")
            )

for density_config in corsika_atmospheres:
    test_densities_cases.append(
        pytest.param(
            "CORSIKA",
            density_config,
            id=f"CORSIKA-{density_config}",
        )
    )

test_densities_cases.append(pytest.param("Isothermal", (None, None), id="Isothermal"))
test_densities_cases.append(
    pytest.param("GeneralizedTarget", (), id="GeneralizedTarget")
)
test_densities_cases.append(
    pytest.param(
        "Unknow",
        (),
        id="Unknown",
        marks=pytest.mark.xfail(reason="Fails for uknown density model"),
    )
)

profiles = {
    "MSIS00": dprof.MSIS00Atmosphere,
    "MSIS00_IC": dprof.MSIS00IceCubeCentered,
    "CORSIKA": dprof.CorsikaAtmosphere,
    "AIRS": dprof.AIRSAtmosphere,
    "Isothermal": dprof.IsothermalAtmosphere,
    "GeneralizedTarget": dprof.GeneralizedTarget,
}


@pytest.mark.parametrize("model, density_config", test_densities_cases)
def test_set_density_profile(mceq_sib21, model, density_config):
    mceq_sib21.set_density_model((model, density_config))
    mceq_sib21.solve()

    # test instances instead of str
    profile = profiles[model](*density_config)
    mceq_sib21.set_density_model(profile)
    mceq_sib21.solve()


def test_set_mod_pprod(mceq_sib21):
    def weight(xmat, egrid, pname, value):
        return (1 + value) * np.ones_like(xmat)

    ret = mceq_sib21.set_mod_pprod(2212, 211, weight, ("a", 0.15))
    assert ret == 1

    assert mceq_sib21._interactions.mod_pprod[(2212, 211)] is not None
    assert mceq_sib21._interactions.mod_pprod[(2212, -211)] is not None

    mceq_sib21.regenerate_matrices()
    mceq_sib21.solve()

    mceq_sib21.regenerate_matrices(skip_decay_matrix=True)
    mceq_sib21.solve()


def test_unset_mod_pprod(mceq_sib21):
    def weight(xmat, egrid, pname, value):
        return (1 + value) * np.ones_like(xmat)

    mceq_sib21.set_mod_pprod(2212, 211, weight, ("a", 0.15))

    mceq_sib21.regenerate_matrices()
    mceq_sib21.solve()

    mceq_sib21.unset_mod_pprod()
    assert not mceq_sib21._interactions.mod_pprod

    mceq_sib21.set_mod_pprod(2212, 211, weight, ("a", 0.15))

    mceq_sib21.regenerate_matrices()
    mceq_sib21.solve()
    mceq_sib21.unset_mod_pprod(dont_fill=False)
    assert not mceq_sib21._interactions.mod_pprod


def test_regenerate_matrices_after_adding_tracking_particle(mceq_sib21):
    mceq_sib21.pman.add_tracking_particle(
        [(2212, 0)], (13, 0), "p_mu", from_interactions=True
    )
    mceq_sib21.regenerate_matrices()
    mceq_sib21.solve()


def test_n_particles_energy_cutoff_and_grid(mceq_sib21):
    import crflux.models as pm

    mceq_sib21.set_primary_model(pm.HillasGaisser2012, "H3a")
    mceq_sib21.solve([0, 1])
    n0 = mceq_sib21.n_particles("mu+", grid_idx=0)
    n1 = mceq_sib21.n_particles("mu+", grid_idx=1)
    n_high_cut = mceq_sib21.n_particles("mu+", grid_idx=1, min_energy_cutoff=1e5)

    assert n0 == 0
    assert n1 > n0
    assert n_high_cut < n1


def test_n_mu_energy_cutoff_and_grid(mceq_sib21):
    import crflux.models as pm

    mceq_sib21.set_primary_model(pm.HillasGaisser2012, "H3a")
    mceq_sib21.solve([0, 1])
    n0 = mceq_sib21.n_mu(grid_idx=0)
    n1 = mceq_sib21.n_mu(grid_idx=1)
    nhigh = mceq_sib21.n_mu(grid_idx=1, min_energy_cutoff=1e5)

    assert n0 == 0
    assert n1 > 0
    assert nhigh < n1


def test_n_e_energy_cutoff_and_grid(mceq_sib21):
    import crflux.models as pm

    mceq_sib21.set_primary_model(pm.HillasGaisser2012, "H3a")
    mceq_sib21.solve([0, 1])
    n0 = mceq_sib21.n_e(grid_idx=0)
    n1 = mceq_sib21.n_e(grid_idx=1)
    nhigh = mceq_sib21.n_e(grid_idx=1, min_energy_cutoff=1e5)

    assert n0 == 0
    assert n1 > 0
    assert nhigh < n1


@pytest.mark.parametrize(
    ["definition", "use_cs_scaling"],
    [
        ["primary_e", False],
        ["no_name_definition", True],
    ],
)
def test_z_factor(mceq_sib21, definition, use_cs_scaling):
    mceq_sib21.solve([0, 1])
    z = mceq_sib21.z_factor(
        2212, 211, definition=definition, use_cs_scaling=use_cs_scaling
    )

    assert isinstance(z, np.ndarray)
    assert z.shape == mceq_sib21.e_grid.shape
    assert np.all(z >= 0)
    assert np.any(z > 0)


def test_decay_z_factor(mceq_sib21):
    mceq_sib21.solve()
    z = mceq_sib21.decay_z_factor(211, 14)

    assert z.shape == mceq_sib21.e_grid.shape
    assert np.any(z > 0)
    assert not np.any(np.isnan(z))


def test_interaction_model_forwarding():
    """Test that user-provided interaction model is correctly forwarded to InteractionCrossSections."""
    import crflux.models as pm

    from MCEq.core import MCEqRun

    # Create MCEqRun with a specific interaction model
    mceq_sib21 = MCEqRun(
        interaction_model="QGSJETII04",
        theta_deg=0.0,
        primary_model=(pm.HillasGaisser2012, "H3a"),
    )

    # Verify that the interaction model is correctly set
    assert mceq_sib21._int_cs.iam == "QGSJETII04"
    assert mceq_sib21._interactions.iam == "QGSJETII04"


def test_ptot_grid(mceq_sib21):
    # Test without bins
    ptot_centers = mceq_sib21.ptot_grid("mu+", return_bins=False)

    assert len(ptot_centers) == len(mceq_sib21.e_grid)
    assert np.all(ptot_centers > 0)
    assert np.all(np.isfinite(ptot_centers))

    # Test with bins
    ptot_bins, ptot_centers_with_bins = mceq_sib21.ptot_grid("mu+", return_bins=True)

    assert len(ptot_bins) == len(mceq_sib21.e_bins)
    assert len(ptot_centers_with_bins) == len(mceq_sib21.e_grid)
    assert np.allclose(ptot_centers, ptot_centers_with_bins)

    # Check that centers are geometric mean of bins
    expected_centers = np.sqrt(ptot_bins[1:] * ptot_bins[:-1])
    assert np.allclose(ptot_centers_with_bins, expected_centers)


def test_etot_grid(mceq_sib21):
    # Test without bins
    etot_centers = mceq_sib21.etot_grid("mu+", return_bins=False)

    assert len(etot_centers) == len(mceq_sib21.e_grid)
    assert np.all(etot_centers > 0)
    assert np.all(np.isfinite(etot_centers))

    # Test with bins
    etot_bins, etot_centers_with_bins = mceq_sib21.etot_grid("mu+", return_bins=True)

    assert len(etot_bins) == len(mceq_sib21.e_bins)
    assert len(etot_centers_with_bins) == len(mceq_sib21.e_grid)
    assert np.allclose(etot_centers, etot_centers_with_bins)

    # Check that bins are kinetic + mass
    mu_mass = mceq_sib21.pman["mu+"].mass
    expected_bins = mceq_sib21.e_bins + mu_mass
    assert np.allclose(etot_bins, expected_bins)


@pytest.mark.parametrize(
    ["return_as", "expected_method"],
    [
        ["kinetic energy", lambda mceq_sib21: (mceq_sib21.e_bins, mceq_sib21.e_grid)],
        [
            "total energy",
            lambda mceq_sib21: mceq_sib21.etot_grid("mu+", return_bins=True),
        ],
        [
            "total momentum",
            lambda mceq_sib21: mceq_sib21.ptot_grid("mu+", return_bins=True),
        ],
    ],
)
@pytest.mark.parametrize("return_bins", [False, True])
def test_xgrid(mceq_sib21, return_as, expected_method, return_bins):
    result = mceq_sib21.xgrid("mu+", return_as, return_bins=return_bins)

    if return_bins:
        bins, centers = result
        expected_bins, expected_centers = expected_method(mceq_sib21)
        assert np.allclose(bins, expected_bins)
        assert np.allclose(centers, expected_centers)
    else:
        expected_bins, expected_centers = expected_method(mceq_sib21)
        assert np.allclose(result, expected_centers)


def test_xgrid_invalid_return_as(mceq_sib21):
    """Test xgrid raises exception for invalid return_as argument."""
    with pytest.raises(Exception, match="Unknown grid type"):
        mceq_sib21.xgrid("mu+", "invalid_type", return_bins=False)


def test_get_set_state_vector_checkpoint_restore(mceq_sib21):
    mceq_sib21.set_primary_model(pm.HillasGaisser2012, "H3a")
    mceq_sib21.set_density_model(dprof.CorsikaAtmosphere("BK_USStd"))

    # First solve
    mceq_sib21.set_theta_deg(0.0)
    mceq_sib21.solve()
    order1, state1 = mceq_sib21._get_state_vector()
    solution1 = mceq_sib21.get_solution("mu+", mag=0, integrate=True)

    # Change something and solve again
    mceq_sib21.set_theta_deg(30.0)
    mceq_sib21.solve()
    order2, state2 = mceq_sib21._get_state_vector()
    solution2 = mceq_sib21.get_solution("mu+", mag=0, integrate=True)

    # Solutions should be different
    assert not np.allclose(solution1, solution2, atol=1e-10)
    assert not np.allclose(state1, state2, atol=1e-14)

    # Restore first state directly (without solving)
    mceq_sib21._solution = np.copy(state1)
    solution1_prime = mceq_sib21.get_solution("mu+", mag=0, integrate=True)

    # Should match original solution
    assert np.allclose(solution1, solution1_prime, atol=1e-10)



"""
Ctypes interface for struct-based interface to the C-version of NRLMSISE-00.
This C version of NRLMSISE-00 is written by Dominik Brodowski
"""

import os
import sysconfig
from ctypes import POINTER, Structure, c_double, c_int, cdll

base = os.path.dirname(os.path.abspath(__file__))
suffix = sysconfig.get_config_var("EXT_SUFFIX")
# Some Python 2.7 versions don't define EXT_SUFFIX
if suffix is None and "SO" in sysconfig.get_config_vars():
    suffix = sysconfig.get_config_var("SO")

assert suffix is not None, "Shared lib suffix was not identified."

for fn in os.listdir(base):
    if "libnrlmsis" in fn and fn.endswith(suffix):
        msis = cdll.LoadLibrary(os.path.join(base, fn))
        break


class nrlmsise_flags(Structure):
    """C-struct containing NRLMSISE related switches"""

    _fields_ = [("switches", c_int * 24), ("sw", c_double * 24), ("swc", c_double * 24)]


class ap_array(Structure):
    """C-struct containing NRLMSISE related switches"""

    _fields_ = [("a", c_double * 7)]


class nrlmsise_input(Structure):
    """The C-struct contains input variables for NRLMSISE."""

    _field_ = [
        ("year", c_int),
        ("doy", c_int),
        ("sec", c_double),
        ("alt", c_double),
        ("g_lat", c_double),
        ("g_long", c_double),
        ("lst", c_double),
        ("f107A", c_double),
        ("f107", c_double),
        ("ap", c_double),
        ("ap_a", POINTER(ap_array)),
    ]


class nrlmsise_output(Structure):
    """The C-struct contains output variables for NRLMSISE."""

    _fields_ = [("d", c_double * 9), ("t", c_double * 2)]



from __future__ import print_function


def test_msis(msis_expected_file):
    from ctypes import byref, c_double, c_int, pointer

    from MCEq.geometry.nrlmsise00 import (
        ap_array,
        msis,
        nrlmsise_flags,
        nrlmsise_input,
        nrlmsise_output,
    )

    with open(msis_expected_file, "r") as f:
        result_expected = f.read()

    output = [nrlmsise_output() for i in range(17)]
    inp = [nrlmsise_input() for i in range(17)]
    flags = nrlmsise_flags()
    aph = ap_array()

    # Inp values
    for i in range(7):
        aph.a[i] = c_double(100.0)

    flags.switches[0] = c_int(0)

    for i in range(1, 24):
        flags.switches[i] = c_int(1)

    for i in range(17):
        inp[i].doy = c_int(172)  # Day of year
        inp[i].year = c_int(0)  # No effect
        inp[i].sec = c_double(29000.0)
        inp[i].alt = c_double(400.0)
        inp[i].g_lat = c_double(60.0)
        inp[i].g_long = c_double(-70.0)
        inp[i].lst = c_double(16.0)
        inp[i].f107A = c_double(150.0)
        inp[i].f107 = c_double(150.0)
        inp[i].ap = c_double(4.0)

    inp[1].doy = c_int(81)
    inp[2].sec = c_double(75000.0)
    inp[2].alt = c_double(1000.0)
    inp[3].alt = c_double(100.0)
    inp[10].alt = c_double(0.0)
    inp[11].alt = c_double(10.0)
    inp[12].alt = c_double(30.0)
    inp[13].alt = c_double(50.0)
    inp[14].alt = c_double(70.0)
    inp[16].alt = c_double(100.0)
    inp[4].g_lat = c_double(0.0)
    inp[5].g_long = c_double(0.0)
    inp[6].lst = c_double(4.0)
    inp[7].f107A = c_double(70.0)
    inp[8].f107 = c_double(180.0)
    inp[9].ap = c_double(40.0)
    inp[15].ap_a = pointer(aph)
    inp[16].ap_a = pointer(aph)
    for i in range(15):
        #     msis.gtd7(byref(inp[i]), byref(flags), byref(output[i]))
        msis.gtd7_py(
            inp[i].year,
            inp[i].doy,
            inp[i].sec,
            inp[i].alt,
            inp[i].g_lat,
            inp[i].g_long,
            inp[i].lst,
            inp[i].f107A,
            inp[i].f107,
            inp[i].ap,
            inp[15].ap_a,
            byref(flags),
            byref(output[i]),
        )
    flags.switches[9] = -1
    for i in range(15, 17):
        msis.gtd7_py(
            inp[i].year,
            inp[i].doy,
            inp[i].sec,
            inp[i].alt,
            inp[i].g_lat,
            inp[i].g_long,
            inp[i].lst,
            inp[i].f107A,
            inp[i].f107,
            inp[i].ap,
            inp[15].ap_a,
            byref(flags),
            byref(output[i]),
        )
    #     msis.gtd7(byref(inp[i]), byref(flags), byref(output[i]))
    # output type 1
    outbuf = ""
    for i in range(17):
        for j in range(9):
            outbuf += f"{output[i].d[j]:E} "
        outbuf += f"{output[i].t[0]:E} "
        outbuf += f"{output[i].t[1]:E} \n"

    # output type 2
    for i in range(3):
        outbuf += "\n"
        outbuf += "\nDAY   "
        for j in range(5):
            outbuf += f"         {inp[i * 5 + j].doy.value:3}"
        outbuf += "\nUT    "
        for j in range(5):
            outbuf += f"       {inp[i * 5 + j].sec.value:5.0f}"
        outbuf += "\nALT   "
        for j in range(5):
            outbuf += f"        {inp[i * 5 + j].alt.value:4.0f}"
        outbuf += "\nLAT   "
        for j in range(5):
            outbuf += f"         {inp[i * 5 + j].g_lat.value:3.0f}"
        outbuf += "\nLONG  "
        for j in range(5):
            outbuf += f"         {inp[i * 5 + j].g_long.value:3.0f}"
        outbuf += "\nLST   "
        for j in range(5):
            outbuf += f"       {inp[i * 5 + j].lst.value:5.0f}"
        outbuf += "\nF107A "
        for j in range(5):
            outbuf += f"         {inp[i * 5 + j].f107A.value:3.0f}"
        outbuf += "\nF107  "
        for j in range(5):
            outbuf += f"         {inp[i * 5 + j].f107.value:3.0f}"
        outbuf += "\n\n"
        outbuf += "\nTINF  "
        for j in range(5):
            outbuf += f"     {output[i * 5 + j].t[0]:7.2f}"
        outbuf += "\nTG    "
        for j in range(5):
            outbuf += f"     {output[i * 5 + j].t[1]:7.2f}"
        outbuf += "\nHE    "
        for j in range(5):
            outbuf += f"   {output[i * 5 + j].d[0]:1.3e}"
        outbuf += "\nO     "
        for j in range(5):
            outbuf += f"   {output[i * 5 + j].d[1]:1.3e}"
        outbuf += "\nN2    "
        for j in range(5):
            outbuf += f"   {output[i * 5 + j].d[2]:1.3e}"
        outbuf += "\nO2    "
        for j in range(5):
            outbuf += f"   {output[i * 5 + j].d[3]:1.3e}"
        outbuf += "\nAR    "
        for j in range(5):
            outbuf += f"   {output[i * 5 + j].d[4]:1.3e}"
        outbuf += "\nH     "
        for j in range(5):
            outbuf += f"   {output[i * 5 + j].d[6]:1.3e}"
        outbuf += "\nN     "
        for j in range(5):
            outbuf += f"   {output[i * 5 + j].d[7]:1.3e}"
        outbuf += "\nANM 0 "
        for j in range(5):
            outbuf += f"   {output[i * 5 + j].d[8]:1.3e}"
        outbuf += "\nRHO   "
        for j in range(5):
            outbuf += f"   {output[i * 5 + j].d[5]:1.3e}"
        outbuf += "\n"
    outbuf += "\n"

    # print(outbuf)

    assert outbuf.strip() == result_expected.strip()



from collections import namedtuple

import numpy as np

from MCEq import config

#: Energy grid (centers, bind widths, dimension)
energy_grid = namedtuple("energy_grid", ("c", "b", "w", "d"))

#: Matrix with x_lab=E_child/E_parent values
_xmat = None

_target_masses = {
    # <A> = 14.6568 (source https://en.wikipedia.org/wiki/Atmosphere_of_Earth)
    "air": sum([f[0] * f[1] for f in [(0.78084, 14), (0.20946, 16), (0.00934, 40)]]),
    "water": 1.0 / 3.0 * (2.0 + 16.0),
    "ice": 1.0 / 3.0 * (2.0 + 16.0),
    "co2": 1.0 / 3.0 * (12.0 + 2.0 * 16.0),
    "rock": 22.0,
    "hydrogen": 1.0,
    "iron": 26.0,
}


def _eval_energy_cuts(e_centers, e_min=None, e_max=None):
    """Evaluate the energy cuts and return the corresponding indices and slice.

    Args:
        e_centers: numpy.ndarray
            Array of energy grid centers.
        e_min: float, optional
            Minimum energy value. Default is None.
        e_max: float, optional
            Maximum energy value. Default is None.

    Returns:
        min_idx: int
            Index corresponding to the minimum energy value.
        max_idx: int
            Index corresponding to the maximum energy value.
        energy_slice: slice
            Slice corresponding to the energy range.

    """
    min_idx, max_idx = 0, len(e_centers)
    energy_slice = slice(None)
    if e_min is not None:
        min_idx = np.argmin(np.abs(e_centers - e_min))
        energy_slice = slice(min_idx, None)
    if e_max is not None:
        max_idx = np.argmin(np.abs(e_centers - e_max)) + 1
        energy_slice = slice(min_idx, max_idx)
    return min_idx, max_idx, energy_slice


def normalize_hadronic_model_name(name):
    """Converts a hadronic model name into a standard form.

    Args:
        name: str
            Hadronic model name.

    Returns:
        str
            Normalized hadronic model name.

    """
    import re

    return re.sub("[-.]", "", name).upper()


def average_A_target(mat="auto"):
    """Average target mass number.

    For air <A> = 14.6568 (using mass fractions from
    https://en.wikipedia.org/wiki/Atmosphere_of_Earth)
    Other media supported are co2, rock, ice, water, and hydrogen.

    Args:
        mat: str or float, optional
            Interaction medium or custom target mass number. Default is "auto".

    Returns:
        float
            Average target mass number.

    Raises:
        ValueError:
            If `mat` is not a valid option.

    """
    if isinstance(mat, str) and mat.lower() == "auto":
        return _target_masses[config.interaction_medium.lower()]
    elif isinstance(mat, str) and mat.lower() in _target_masses:
        return _target_masses[mat.lower()]
    elif isinstance(mat, float) or isinstance(mat, int):
        return float(mat)
    else:
        raise ValueError(
            "mceq_config.A_target is expected to be a "
            + 'number or one of {0} or "auto"'.format(", ".join(_target_masses.keys()))
        )


def theta_deg(cos_theta):
    """Converts :math:`\\cos{\\theta}` to :math:`\\theta` in degrees."""
    return np.rad2deg(np.arccos(cos_theta))


def gen_xmat(energy_grid):
    """Generates x_lab matrix for a given energy grid"""
    global _xmat
    dims = (energy_grid.d, energy_grid.d)
    if _xmat is None or _xmat.shape != dims:
        _xmat = np.zeros(dims)
        for eidx in range(energy_grid.d):
            xvec = energy_grid.c[: eidx + 1] / energy_grid.c[eidx]
            _xmat[: eidx + 1, eidx] = xvec
    return _xmat


def print_in_rows(min_dbg_level, str_list, n_cols=5):
    """Prints contents of a list in rows `n_cols`
    entries per row.
    """
    if min_dbg_level > config.debug_level:
        return

    ls = len(str_list)
    n_full_length = int(ls / n_cols)
    n_rest = ls % n_cols
    print_str = "\n"
    for i in range(n_full_length):
        print_str += ('"{:}", ' * n_cols).format(
            *str_list[i * n_cols : (i + 1) * n_cols]
        ) + "\n"
    print_str += ('"{:}", ' * n_rest).format(*str_list[-n_rest:])

    print(print_str.strip()[:-1])


def is_charm_pdgid(pdgid):
    """Returns True if particle ID belongs to a heavy (charm) hadron."""

    return (abs(pdgid) > 400 and abs(pdgid) < 500) or (
        abs(pdgid) > 4000 and abs(pdgid) < 5000
    )


def _get_closest(value, in_list):
    """Returns the closes value to 'value' from given list."""

    minindex = np.argmin(np.abs(in_list - value * np.ones(len(in_list))))
    return minindex, in_list[minindex]


def getAZN(pdg_id):
    """Returns mass number :math:`A`, charge :math:`Z` and neutron
    number :math:`N` of ``pdg_id``.

    Note::

        PDG ID for nuclei is coded according to 10LZZZAAAI.
        For iron-52 it is 1000260520.

    Args:
        pdgid (int): PDG ID of nucleus/mass group
    Returns:
        (int,int,int): (Z,A) tuple
    """
    Z, A = 1, 1
    if pdg_id < 2000:
        return 0, 0, 0
    if pdg_id == 2112:
        return 1, 0, 1
    if pdg_id == 2212:
        return 1, 1, 0
    if pdg_id > 1000000000:
        A = pdg_id % 1000 / 10
        Z = pdg_id % 1000000 / 10000
        return A, Z, A - Z
    return 1, 0, 0


def getAZN_corsika(corsikaid):
    """Returns mass number :math:`A`, charge :math:`Z` and neutron
    number :math:`N` of ``corsikaid``.

    Args:
        corsikaid (int): corsika id of nucleus/mass group
    Returns:
        (int,int,int): (Z,A) tuple
    """
    Z, A = 1, 1

    if corsikaid == 14:
        return getAZN(2212)
    if corsikaid >= 100:
        Z = corsikaid % 100
        A = (corsikaid - Z) // 100
    else:
        Z, A = 0, 0

    return A, Z, A - Z


def corsikaid2pdg(corsika_id):
    """Conversion of CORSIKA nuclear code to PDG nuclear code"""
    if corsika_id in [101, 14]:
        return 2212
    if corsika_id in [100, 13]:
        return 2112
    A, Z, _ = getAZN_corsika(corsika_id)
    # 10LZZZAAAI
    pdg_id = 1000000000
    pdg_id += 10 * A
    pdg_id += 10000 * Z
    return pdg_id


def pdg2corsikaid(pdg_id):
    """Conversion from nuclear PDG ID to CORSIKA ID.

    Note::

        PDG ID for nuclei is coded according to 10LZZZAAAI.
        For iron-52 it is 1000260520.
    """
    if pdg_id == 2212:
        return 14

    A = pdg_id % 1000 / 10
    Z = pdg_id % 1000000 / 10000

    return A * 100 + Z


def caller_name(skip=2):
    """Get a name of a caller in the format module.class.method

    `skip` specifies how many levels of stack to skip while getting caller
    name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.
    An empty string is returned if skipped levels exceed stack height.abs

    From https://gist.github.com/techtonik/2151727
    """
    import inspect

    stack = inspect.stack()
    start = 0 + skip

    if len(stack) < start + 1:
        return ""

    parentframe = stack[start][0]

    name = []

    if config.print_module:
        module = inspect.getmodule(parentframe)
        # `modname` can be None when frame is executed directly in console
        if module:
            name.append(module.__name__ + ".")

    # detect classname
    if "self" in parentframe.f_locals:
        # I don't know any way to detect call from the object method
        # there seems to be no way to detect static method call - it will
        # be just a function call

        name.append(parentframe.f_locals["self"].__class__.__name__ + "::")

    codename = parentframe.f_code.co_name
    if codename != "<module>":  # top level usually
        name.append(codename + "(): ")  # function or a method
    else:
        name.append(": ")  # If called from module scope

    del parentframe
    return "".join(name)


def info(min_dbg_level, *message, **kwargs):
    """Print to console if `min_debug_level <= config.debug_level`

    The fuction determines automatically the name of caller and appends
    the message to it. Message can be a tuple of strings or objects
    which can be converted to string using `str()`.

    Args:
        min_dbg_level (int): Minimum debug level in config for printing
        message (tuple): Any argument or list of arguments that casts to str
        condition (bool): Print only if condition is True
        blank_caller (bool): blank the caller name (for multiline output)
        no_caller (bool): don't print the name of the caller

    Authors:
        Anatoli Fedynitch (DESY)
        Jonas Heinze (DESY)
    """
    condition = kwargs.pop("condition", True)
    blank_caller = kwargs.pop("blank_caller", False)
    no_caller = kwargs.pop("no_caller", False)
    if config.override_debug_fcn and min_dbg_level < config.override_max_level:
        fcn_name = caller_name(skip=2).split("::")[-1].split("():")[0]
        if fcn_name in config.override_debug_fcn:
            min_dbg_level = 0

    if condition and min_dbg_level <= config.debug_level:
        message = [str(m) for m in message]
        cname = caller_name() if not no_caller else ""
        if blank_caller:
            cname = len(cname) * " "
        print(cname + " ".join(message))



import sys
import warnings

# Import the actual configuration module
from MCEq import config as _config

# Warn users about the deprecated module path
warnings.warn(
    "The module 'mceq_config' has been moved to 'MCEq.config'. "
    "Please update your `import mceq_config` to `import MCEq.config` accordingly.",
    DeprecationWarning,
    stacklevel=2,
)

# Replace this module in sys.modules so that any attribute access
# or subsequent imports refer to the real config module
sys.modules[__name__] = _config



from collections import defaultdict
from os.path import isfile, join

import h5py
import numpy as np
import six

from MCEq import config

from .misc import _eval_energy_cuts, info, normalize_hadronic_model_name

# TODO: Convert this to some functional generic class. Very erro prone to
# enter stuff by hand
equivalences = {
    "SIBYLL23": {
        -4132: 4122,
        -4122: 4122,
        -3334: -3312,
        -3322: -2112,
        -3212: -3122,
        -413: -411,
        113: 211,
        221: 211,
        111: 211,
        310: 130,
        413: 411,
        3212: 3122,
        3334: 3312,
    },
    "SIBYLL21": {
        -3322: 2112,
        -3312: 2212,
        -3222: 2212,
        -3212: 2112,
        -3122: 2112,
        -3112: 2212,
        -2212: 2212,
        -2112: 2112,
        310: 130,
        111: 211,
        3112: 2212,
        3122: 2112,
        3212: 2112,
        3222: 2212,
        3312: 2212,
        3322: 2112,
    },
    "QGSJET01": {
        -4122: 2212,
        -3322: 2212,
        -3312: 2212,
        -3222: 2212,
        -3122: 2212,
        -3112: 2212,
        -2212: 2212,
        -2112: 2212,
        -421: 321,
        -411: 321,
        -211: 211,
        -321: 321,
        111: 211,
        221: 211,
        130: 321,
        310: 321,
        411: 321,
        421: 321,
        2112: 2212,
        3112: 2212,
        3122: 2212,
        3222: 2212,
        3312: 2212,
        3322: 2212,
        4122: 2212,
    },
    "QGSJETII": {
        -3122: -2112,
        111: 211,
        113: 211,
        221: 211,
        310: 130,
        3122: 2112,
    },
    "DPMJET": {
        -4122: -3222,
        -3334: -3312,
        -3212: -3122,
        -431: -321,
        -421: -321,
        -413: -321,
        -411: -321,
        310: 130,
        113: 211,
        221: 211,
        111: 211,
        411: 321,
        413: 321,
        421: 321,
        431: 321,
        3212: 3122,
        3334: 3312,
        4122: 3222,
    },
    "EPOSLHC": {
        -3334: 2212,
        -3322: -3122,
        -3312: 2212,
        -3222: -2212,
        -3212: -3122,
        -3112: 2212,
        111: 211,
        113: 211,
        221: 211,
        310: 130,
        3112: -2212,
        3212: 3122,
        3222: 2212,
        3312: -2212,
        3322: 3122,
        3334: -2212,
    },
    "PYTHIA8": {
        -3122: -2112,
        -431: -321,
        -421: -321,
        -413: -321,
        -411: -321,
        111: 211,
        113: 211,
        221: 211,
        310: 321,
        130: 321,
        411: 321,
        413: 321,
        421: 321,
        431: 321,
        3122: 2112,
    },
}

equivalences["FLUKA"] = equivalences["DPMJET"]


class HDF5Backend:
    """Provides access to tabulated data stored in an HDF5 file.

    The file contains all necessary ingredients to run MCEq, i.e. no
    other files are required. This database is not maintained in git
    and it will change infrequently.
    """

    def __init__(self, medium=config.interaction_medium):
        info(2, "Opening HDF5 file", config.mceq_db_fname)
        self.had_fname = join(config.data_dir, config.mceq_db_fname)
        if not isfile(self.had_fname):
            raise Exception(
                f'MCEq DB file {config.mceq_db_fname} not found in "data" directory.'
            )

        self.em_fname = join(config.data_dir, config.em_db_fname)
        if config.enable_em and not isfile(self.had_fname):
            _n = config.em_db_fname
            raise Exception(
                f'Electromagnetic DB file {_n} not found in "data" directory.'
            )

        with h5py.File(self.had_fname, "r") as mceq_db:
            from MCEq.misc import energy_grid

            ca = mceq_db["common"].attrs
            self.version = (
                mceq_db.attrs["version"] if "version" in mceq_db.attrs else "1.0.0"
            )
            self.min_idx, self.max_idx, self._cuts = _eval_energy_cuts(
                ca["e_grid"], config.e_min, config.e_max
            )
            self._energy_grid = energy_grid(
                ca["e_grid"][self._cuts],
                ca["e_bins"][self.min_idx : self.max_idx + 1],
                ca["widths"][self._cuts],
                int(self.max_idx - self.min_idx),
            )
            self.dim_full = int(ca["e_dim"])

        self.medium = medium

    @property
    def energy_grid(self):
        return self._energy_grid

    def _gen_db_dictionary(self, hdf_root, indptrs, equivalences={}):
        from scipy.sparse import csr_matrix

        index_d = {}
        relations = defaultdict(list)
        particle_list = []
        if "description" in hdf_root.attrs:
            description = hdf_root.attrs["description"]
        else:
            description = None
        mat_data = np.asarray(hdf_root[:, :], dtype=config.floatlen)
        indptr_data = np.asarray(indptrs[:])
        len_data = hdf_root.attrs["len_data"]
        if hdf_root.attrs["tuple_idcs"].shape[1] == 4:
            model_particles = sorted(
                list(set(hdf_root.attrs["tuple_idcs"][:, (0, 2)].flatten().tolist()))
            )
        else:
            model_particles = sorted(
                list(set(hdf_root.attrs["tuple_idcs"].flatten().tolist()))
            )

        exclude = config.adv_set["disabled_particles"]
        read_idx = 0
        available_parents = [
            (pdg, parity) for (pdg, parity) in (hdf_root.attrs["tuple_idcs"][:, :2])
        ]
        available_parents = sorted(list(set(available_parents)))

        # Reverse equivalences
        eqv_lookup = defaultdict(list)
        for k in equivalences:
            eqv_lookup[(equivalences[k], 0)].append((k, 0))

        for tupidx, tup in enumerate(hdf_root.attrs["tuple_idcs"]):
            if len(tup) == 4:
                parent_pdg, child_pdg = tuple(tup[:2]), tuple(tup[2:])
            elif len(tup) == 2:
                parent_pdg, child_pdg = (tup[0], 0), (tup[1], 0)
            else:
                raise Exception("Failed decoding parent-child relation.")

            if (abs(parent_pdg[0]) in exclude) or (abs(child_pdg[0]) in exclude):
                read_idx += len_data[tupidx]
                continue
            parent_pdg = int(parent_pdg[0]), (parent_pdg[1])
            child_pdg = int(child_pdg[0]), (child_pdg[1])

            particle_list.append(parent_pdg)
            particle_list.append(child_pdg)
            index_d[(parent_pdg, child_pdg)] = np.asarray(
                (
                    csr_matrix(
                        (
                            mat_data[0, read_idx : read_idx + len_data[tupidx]],
                            mat_data[1, read_idx : read_idx + len_data[tupidx]],
                            indptr_data[tupidx, :],
                        ),
                        shape=(self.dim_full, self.dim_full),
                    )[self._cuts, self.min_idx : self.max_idx]
                ).toarray()
            )
            relations[parent_pdg].append(child_pdg)

            info(
                20,
                f"This parent {parent_pdg[0]} is used for interactions of",
                [p[0] for p in eqv_lookup[parent_pdg]],
                condition=len(equivalences) > 0,
            )
            if config.assume_nucleon_interactions_for_exotics:
                for eqv_parent in eqv_lookup[parent_pdg]:
                    if eqv_parent[0] not in model_particles:
                        info(
                            10,
                            "No equiv. replacement needed of",
                            eqv_parent,
                            "for",
                            parent_pdg,
                            "parent.",
                        )
                        continue
                    elif eqv_parent in available_parents:
                        info(
                            10,
                            f"Parent {eqv_parent[0]} has dedicated simulation.",
                        )
                        continue
                    particle_list.append(eqv_parent)
                    index_d[(eqv_parent, child_pdg)] = index_d[(parent_pdg, child_pdg)]
                    relations[eqv_parent] = relations[parent_pdg]
                    _e = eqv_parent[0]
                    _p = parent_pdg[0]
                    info(
                        15,
                        f"equivalence of {_e} and {_p} interactions",
                    )

            read_idx += len_data[tupidx]

        return {
            "parents": sorted(list(relations)),
            "particles": sorted(list(set(particle_list))),
            "relations": dict(relations),
            "index_d": dict(index_d),
            "description": description,
        }

    def _check_subgroup_exists(self, subgroup, mname):
        available_models = [m for m in list(subgroup) if "indptrs" not in m]
        if mname not in available_models:
            info(0, "Invalid choice/model", mname)
            info(0, "Choose from:\n", "\n".join(available_models))
            raise Exception("Unknown selections.")

    def interaction_db(self, interaction_model_name):
        mname = normalize_hadronic_model_name(interaction_model_name)
        info(10, f"Generating interaction db. mname={mname}")
        with h5py.File(self.had_fname, "r") as mceq_db:
            if (
                self.medium not in mceq_db["hadronic_interactions"]
                or mname not in mceq_db["hadronic_interactions"][self.medium]
            ) and config.fallback_to_air_cs:
                self._check_subgroup_exists(
                    mceq_db["hadronic_interactions"]["air"], mname
                )
                info(
                    1,
                    (
                        f"Production matrices for {mname} in {self.medium} not found."
                        + "Fall-back to air."
                    ),
                )
                medium = "air"
            else:
                self._check_subgroup_exists(
                    mceq_db["hadronic_interactions"][self.medium], mname
                )
                medium = self.medium

            if "SIBYLL21" in mname:
                eqv = equivalences["SIBYLL21"]
            elif "SIBYLL23" in mname:
                eqv = equivalences["SIBYLL23"]
            elif "QGSJET01" in mname:
                eqv = equivalences["QGSJET01"]
            elif "QGSJETII" in mname:
                eqv = equivalences["QGSJETII"]
            elif "DPMJET" in mname:
                eqv = equivalences["DPMJET"]
            elif "EPOSLHC" in mname:
                eqv = equivalences["EPOSLHC"]
            elif "PYTHIA8" in mname:
                eqv = equivalences["PYTHIA8"]
            elif "FLUKA" in mname:
                eqv = equivalences["FLUKA"]
            else:
                raise ValueError("Unknown equivalence table for", mname)
            int_index = self._gen_db_dictionary(
                mceq_db["hadronic_interactions"][medium][mname],
                mceq_db["hadronic_interactions"][medium][mname + "_indptrs"],
                equivalences=eqv,
            )

        # Append electromagnetic interaction matrices from EmCA
        if config.enable_em:
            if medium == "ice":
                info(5, "Electromagnetic cross sections for ice replaced by water.")
                medium = "water"

            with h5py.File(self.em_fname, "r") as em_db:
                info(2, "Injecting EmCA matrices into interaction_db.")
                self._check_subgroup_exists(em_db, "electromagnetic")
                self._check_subgroup_exists(em_db["electromagnetic"], self.medium)
                em_index = self._gen_db_dictionary(
                    em_db["electromagnetic"][self.medium]["emca_mats"],
                    em_db["electromagnetic"][self.medium]["emca_mats" + "_indptrs"],
                )
            if config.muon_helicity_dependence:
                # This is only approximately valid and is done for consistency.
                # Typically electrons would quickly depolarize due to multiple
                # scattering but this requires additional matrices for
                # (-11,1) -> (-11,0) etc. that are not available now.
                from itertools import product

                info(
                    5,
                    "Copy bremsstrahlung and photon emission to "
                    + "polarised electrons and muons.",
                )
                for pid, h in product([11, -11, 13, -13], [-1, 1]):
                    em_index["index_d"][((pid, h), (pid, h))] = em_index["index_d"][
                        ((pid, 0), (pid, 0))
                    ]
                    em_index["index_d"][((pid, h), (22, 0))] = em_index["index_d"][
                        ((pid, 0), (22, 0))
                    ]
                    em_index["parents"].append((pid, h))

                em_index["relations"] = defaultdict(list)
                em_index["particles"] = []

                for idx_tup in em_index["index_d"]:
                    parent, child = idx_tup
                    em_index["relations"][parent].append(child)
                    em_index["particles"].append(parent)
                    em_index["particles"].append(child)

            int_index["parents"] = sorted(int_index["parents"] + em_index["parents"])
            int_index["particles"] = sorted(
                list(set(int_index["particles"] + em_index["particles"]))
            )
            int_index["relations"].update(em_index["relations"])
            int_index["index_d"].update(em_index["index_d"])

        if int_index["description"] is not None:
            int_index["description"] += "\nInteraction model name: " + mname
        else:
            int_index["description"] = "Interaction model name: " + mname

        return int_index

    def decay_db(self, decay_dset_name):
        info(10, f"Generating decay db. dset_name={decay_dset_name}")

        with h5py.File(self.had_fname, "r") as mceq_db:
            if config.muon_helicity_dependence:
                if decay_dset_name != "polarized":
                    info(
                        0,
                        "Warning: "
                        + f"Does this decay dataset '{decay_dset_name}'"
                        + " include polarization?",
                    )
                else:
                    decay_dset_name = "polarized"
                info(2, "Using helicity dependent decays.")

            self._check_subgroup_exists(mceq_db["decays"], decay_dset_name)
            dec_index = self._gen_db_dictionary(
                mceq_db["decays"][decay_dset_name],
                mceq_db["decays"][decay_dset_name + "_indptrs"],
            )

            # Refresh the metadata after modifying the index
            dec_index["relations"] = defaultdict(list)
            dec_index["particles"] = []

            for idx_tup in dec_index["index_d"]:
                parent, child = idx_tup
                dec_index["relations"][parent].append(child)
                dec_index["particles"].append(parent)
                dec_index["particles"].append(child)

            dec_index["parents"] = sorted(list(dec_index["relations"]))
            dec_index["particles"] = sorted(list(set(dec_index["particles"])))
        return dec_index

    def cs_db(self, interaction_model_name):
        mname = normalize_hadronic_model_name(interaction_model_name)
        medium = self.medium
        if "SIBYLL23C" in mname or "SIBYLL23DSTAR" in mname:
            info(5, f"{mname} cross sections replaced by 23D.")
            mname = "SIBYLL23D"

        if "FLUKA" in mname:
            info(5, f"{mname} cross sections replaced by DPMIII191.")
            mname = "DPMJETIII191"

        if config.adv_set["forced_int_cs"] is not None:
            mname = config.adv_set["forced_int_cs"]
            info(1, "All interaction cross sections forced to", mname)

        if medium == "air-legacy" and "SIBYLL23" not in mname:
            info(5, "air-legacy target replaced by air for", mname)
            medium = "air"

        with h5py.File(self.had_fname, "r") as mceq_db:
            self._check_subgroup_exists(mceq_db["cross_sections"], medium)
            self._check_subgroup_exists(mceq_db["cross_sections"][medium], mname)
            cs_db = mceq_db["cross_sections"][medium][mname]
            cs_data = cs_db[:]
            index_d = {}
            parents = list(cs_db.attrs["parents"])
            for ip, p in enumerate(parents):
                index_d[p] = cs_data[self._cuts, ip]

        if config.adv_set["replace_meson_cross_sections_with"] is not None:
            mname_mesons = config.adv_set["replace_meson_cross_sections_with"]
            info(1, "Meson cross sections forced to", mname_mesons)
            with h5py.File(self.had_fname, "r") as mceq_db:
                self._check_subgroup_exists(mceq_db["cross_sections"], medium)
                self._check_subgroup_exists(
                    mceq_db["cross_sections"][medium], mname_mesons
                )
                mes_cs_db = mceq_db["cross_sections"][medium][mname_mesons]
                mes_cs_data = mes_cs_db[:]
                mes_parents = list(mes_cs_db.attrs["projectiles"])
                for ip, p in enumerate(mes_parents):
                    if p in index_d and (100 < abs(p) < 2000):
                        info(1, "Meson cross sections for", p, "replaced.")
                        index_d[p] = mes_cs_data[self._cuts, ip]

        # Append electromagnetic interaction cross sections from EmCA
        if config.enable_em:
            with h5py.File(self.em_fname, "r") as em_db:
                info(2, "Injecting EmCA matrices into interaction_db.")
                self._check_subgroup_exists(em_db, "electromagnetic")
                self._check_subgroup_exists(em_db["electromagnetic"], medium)
                em_cs = em_db["electromagnetic"][medium]["cs"][:]
                em_parents = list(
                    em_db["electromagnetic"][medium]["cs"].attrs["projectiles"]
                )

                for ip, p in enumerate(em_parents):
                    if p in index_d:
                        raise Exception("EM cross sections already in database?")
                    index_d[p] = em_cs[ip, self._cuts]
                parents += em_parents

        return {"parents": parents, "index_d": index_d}

    def continuous_loss_db(self):
        with h5py.File(self.had_fname, "r") as mceq_db:
            self._check_subgroup_exists(mceq_db["continuous_losses"], self.medium)
            if config.enable_em or not config.enable_cont_rad_loss:
                loss_case = "ionization"
            else:
                loss_case = "total"
            cl_db = mceq_db["continuous_losses"][self.medium][loss_case]
            # No rad losses for hadrons implememented
            cl_db_hadrons = mceq_db["continuous_losses"][self.medium]["total"]
            index_d = {}
            generic_dedx = None

            for k in list(cl_db):
                if k != "hadron":
                    for hel in [0, 1, -1]:
                        index_d[(int(k), hel)] = cl_db[k][self._cuts]
                else:
                    # Tuple (boost, dEdx)
                    generic_dedx = (cl_db_hadrons[k][0], cl_db_hadrons[k][1])

            # if config.enable_em:
            #     with h5py.File(self.em_fname, "r") as em_db:
            #         info(2, "Injecting EmCA matrices into interaction_db.")
            #         self._check_subgroup_exists(em_db, "electromagnetic")
            #         for hel in [0, 1, -1]:
            #             index_d[(11, hel)] = em_db["electromagnetic"][self.medium][
            #                 "dEdX 11"
            #             ][self._cuts]
            #             index_d[(-11, hel)] = em_db["electromagnetic"][self.medium][
            #                 "dEdX -11"
            #             ][self._cuts]
        if generic_dedx is not None:
            return {
                "parents": sorted(list(index_d)),
                "index_d": index_d,
                "generic": generic_dedx,
            }
        else:
            return {"parents": sorted(list(index_d)), "index_d": index_d}


class Interactions:
    """Class for managing the dictionary of interaction yield matrices.

    Args:
    mceq_hdf_db (object): instance of :class:`MCEq.data.HDF5Backend`
    """

    def __init__(self, mceq_hdf_db):
        from collections import defaultdict

        #: MCEq HDF5Backend reference
        self.mceq_db = mceq_hdf_db
        #: reference to energy grid
        self.energy_grid = mceq_hdf_db.energy_grid
        #: List of active parents
        self.parents = None
        #: List of all known particles
        self.particles = None
        #: Dictionary parent/child relations
        self.relations = None
        #: Dictionary containing the distribuiton matrices
        self.index_d = None
        #: String containing the desciption of the model
        self.description = None

        #: (str) Interaction Model name
        self.iam = None
        # #: (tuple) selection of a band of coeffictients (in xf)
        # self.band = None
        #: (tuple) modified particle combination for error prop.
        self.mod_pprod = defaultdict(dict)

    def load(self, interaction_model, parent_list=None):
        from MCEq.misc import is_charm_pdgid

        self.iam = normalize_hadronic_model_name(interaction_model)
        # Load tables and index from file
        index = self.mceq_db.interaction_db(self.iam)
        disabled_particles = config.adv_set["disabled_particles"]
        self.parents = [p for p in index["parents"] if p[0] not in disabled_particles]
        self.relations = index["relations"]
        self.index_d = index["index_d"]
        self.description = index["description"]

        # Advanced options

        if parent_list is not None:
            self.parents = [
                p
                for p in self.parents
                if p in parent_list and p[0] not in disabled_particles
            ]
        if config.adv_set["disable_charm_pprod"]:
            self.parents = [p for p in self.parents if not is_charm_pdgid(p[0])]
        if config.adv_set["disable_interactions_of_unstable"]:
            self.parents = [
                p for p in self.parents if p[0] not in [2212, 2112, -2212, -2112]
            ]
        if config.adv_set["allowed_projectiles"]:
            self.parents = [
                p for p in self.parents if p[0] in config.adv_set["allowed_projectiles"]
            ]

        self.particles = []
        for p in list(self.relations):
            if p not in self.parents:
                _ = self.relations.pop(p, None)
                continue
            self.particles.append(p)
            self.particles += [
                d for d in self.relations[p] if d not in disabled_particles
            ]
        self.particles = sorted(list(set(self.particles)))
        if config.adv_set["disable_direct_leptons"]:
            for p in list(self.relations):
                self.relations[p] = [
                    c for c in self.relations[p] if not 10 < abs(c[0]) < 20
                ]

        if len(disabled_particles) > 0:
            for p in list(self.relations):
                self.relations[p] = [
                    c for c in self.relations[p] if c[0] not in disabled_particles
                ]
        if not self.particles:
            info(
                2, "None of the parent_list particles interact. Returning custom list."
            )
            self.particles = parent_list

    def __getitem__(self, key):
        return self.get_matrix(*key)

    def __contains__(self, key):
        """Defines the `in` operator to look for particles"""
        return key in self.parents

    def _gen_mod_matrix(self, x_func, *args):
        """Creates modification matrix using an (x,E)-dependent function.

        :math:`x = \\frac{E_{\\rm primary}}{E_{\\rm secondary}}` is the
        fraction of secondary particle energy. ``x_func`` can be an
        arbitrary function modifying the :math:`x_\\text{lab}` distribution.
        Run this method each time you change ``x_func``, or its parameters,
        not each time you change modified particle.
        The ``args`` are passed to the function.

        Args:
          x_func (object): reference to function
          args (tuple): arguments of `x_func`

        Returns:
          (numpy.array): modification matrix
        """
        from MCEq.misc import gen_xmat

        info(2, "Generating modification matrix for", x_func.__name__, args)

        xmat = gen_xmat(self.energy_grid)

        # select the relevant slice of interaction matrix
        modmat = x_func(xmat, self.energy_grid.c, *args)
        # Set lower triangular indices to 0. (should be not necessary)
        modmat[np.tril_indices(self.energy_grid.d, -1)] = 0.0

        return np.asarray(modmat)

    def _set_mod_pprod(self, prim_pdg, sec_pdg, x_func, args):
        """Sets combination of parent/secondary for error propagation.

        The production spectrum of ``sec_pdg`` in interactions of
        ``prim_pdg`` is modified according to the function passed to
        :func:`InteractionYields.init_mod_matrix`

        Args:
          prim_pdg (int): interacting (primary) particle PDG ID
          sec_pdg (int): secondary particle PDG ID
        """

        # Short cut for the pprod list
        mpli = self.mod_pprod
        pstup = (prim_pdg, sec_pdg)

        if config.use_isospin_sym and prim_pdg not in [2212, 2112]:
            raise Exception("Unsupported primary for isospin symmetries.")

        if (x_func.__name__, args) in mpli[(pstup)]:
            info(
                5,
                " no changes to particle production"
                + f" modification matrix of {prim_pdg}/{sec_pdg} for {x_func.__name__},{args}",
            )
            return False

        # Check function with same mode but different parameter is supplied
        for xf_name, fargs in list(mpli[pstup]):
            if (xf_name == x_func.__name__) and (fargs[0] == args[0]):
                info(
                    1,
                    "Warning. If you modify only the value of a function,",
                    "unset and re-apply all changes",
                )
                return False

        info(
            2,
            "modifying modify particle production"
            + f" matrix of {prim_pdg}/{sec_pdg} for {x_func.__name__},{args}",
        )

        kmat = self._gen_mod_matrix(x_func, *args)
        mpli[pstup][(x_func.__name__, args)] = kmat

        info(5, 'modification "strength"', np.sum(kmat) / np.count_nonzero(kmat))

        if not config.use_isospin_sym:
            return True

        prim_pdg, symm_pdg = 2212, 2112
        if prim_pdg == 2112:
            prim_pdg = 2112
            symm_pdg = 2212

        # p->pi+ = n-> pi-, p->pi- = n-> pi+
        if abs(sec_pdg) == 211:
            # Add the same mod to the isospin symmetric particle combination
            mpli[(symm_pdg, -sec_pdg)][("isospin", args)] = kmat

            # Assumption: Unflavored production coupled to the average
            # of pi+ and pi- production

            if np.any([p in self.parents for p in [221, 223, 333]]):
                unflv_arg = None
                if (prim_pdg, -sec_pdg) not in mpli:
                    # Only pi+ or pi- (not both) have been modified
                    unflv_arg = (args[0], 0.5 * args[1])

                if (prim_pdg, -sec_pdg) in mpli:
                    # Compute average of pi+ and pi- modification matrices
                    # Save the 'average' argument (just for meaningful output)
                    for arg_name, arg_val in mpli[(prim_pdg, -sec_pdg)]:
                        if arg_name == args[0]:
                            unflv_arg = (args[0], 0.5 * (args[1] + arg_val))

                unflmat = self._gen_mod_matrix(x_func, *unflv_arg)

                # modify eta, omega, phi, 221, 223, 333
                for t in [
                    (prim_pdg, 221),
                    (prim_pdg, 223),
                    (prim_pdg, 333),
                    (symm_pdg, 221),
                    (symm_pdg, 223),
                    (symm_pdg, 333),
                ]:
                    mpli[t][("isospin", unflv_arg)] = unflmat

        # Charged and neutral kaons
        elif abs(sec_pdg) == 321:
            # approx.: p->K+ ~ n-> K+, p->K- ~ n-> K-
            mpli[(symm_pdg, sec_pdg)][("isospin", args)] = kmat
            k0_arg = (args[0], 0.5 * args[1])
            if (prim_pdg, -sec_pdg) in mpli:
                # Compute average of K+ and K- modification matrices
                # Save the 'average' argument (just for meaningful printout)
                for arg_name, arg_val in mpli[(prim_pdg, -sec_pdg)]:
                    if arg_name == args[0]:
                        k0_arg = (args[0], 0.5 * (args[1] + arg_val))

            k0mat = self._gen_mod_matrix(x_func, *k0_arg)

            # modify K0L/S
            for t in [
                (prim_pdg, 310),
                (prim_pdg, 130),
                (symm_pdg, 310),
                (symm_pdg, 130),
            ]:
                mpli[t][("isospin", k0_arg)] = k0mat

        elif abs(sec_pdg) == 411:
            ssec = np.sign(sec_pdg)
            mpli[(prim_pdg, ssec * 421)][("isospin", args)] = kmat
            mpli[(prim_pdg, ssec * 431)][("isospin", args)] = kmat
            mpli[(symm_pdg, sec_pdg)][("isospin", args)] = kmat
            mpli[(symm_pdg, ssec * 421)][("isospin", args)] = kmat
            mpli[(symm_pdg, ssec * 431)][("isospin", args)] = kmat

        # Leading particles
        elif abs(sec_pdg) == prim_pdg:
            mpli[(symm_pdg, symm_pdg)][("isospin", args)] = kmat
        elif abs(sec_pdg) == symm_pdg:
            mpli[(symm_pdg, prim_pdg)][("isospin", args)] = kmat
        else:
            info(0, " Warning: No isospin relation found for secondary" + str(sec_pdg))

        # Tell MCEqRun to regenerate the matrices if something has changed
        return True

    def print_mod_pprod(self):
        """Prints the active particle production modification."""

        for i, (prim_pdg, sec_pdg) in enumerate(sorted(self.mod_pprod)):
            for j, (argname, argv) in enumerate(self.mod_pprod[(prim_pdg, sec_pdg)]):
                info(
                    2,
                    f"{i + j}: {prim_pdg} -> {sec_pdg}, func: {argname}, arg: {argv}",
                    no_caller=True,
                )

    def get_matrix(self, parent, child):
        """Returns a ``DIM x DIM`` yield matrix.

        Args:
          parent (int): PDG ID of parent particle
          child (int): PDG ID of final state child/secondary particle
        Returns:
          numpy.array: yield matrix
        """
        info(10, "Called for", parent, child)
        if child not in self.relations[parent]:
            raise Exception(f"trying to get empty matrix {parent} -> {child}")

        m = self.index_d[(parent, child)]

        if (
            config.adv_set["disable_leading_mesons"]
            and abs(child) < 2000
            and (parent, -child) in list(self.index_d)
        ):
            m_anti = self.index_d[(parent, -child)]
            ie = 50
            info(
                5,
                "sum in disable_leading_mesons",
                (np.sum(m[:, ie - 30 : ie]) - np.sum(m_anti[:, ie - 30 : ie])),
            )

            if (np.sum(m[:, ie - 30 : ie]) - np.sum(m_anti[:, ie - 30 : ie])) > 0:
                info(
                    5,
                    "inverting meson due to leading particle veto.",
                    child,
                    "->",
                    -child,
                )
                m = m_anti
            else:
                info(5, "no inversion since child not leading", child)
        else:
            info(20, "no meson inversion in leading particle veto.", parent, child)
        if (parent[0], child[0]) in list(self.mod_pprod):
            info(
                5,
                f"using modified particle production for {parent[0]}/{child[0]}",
            )
            i = 0
            m = np.copy(m)
            for args, mmat in six.iteritems(self.mod_pprod[(parent[0], child[0])]):
                info(10, i, (parent[0], child[0]), args, np.sum(mmat), np.sum(m))
                i += 1
                m *= mmat

        return m


class Decays:
    """Class for managing the dictionary of decay yield matrices.

    Args:
      mceq_hdf_db (object): instance of :class:`MCEq.data.HDF5Backend`
    """

    def __init__(self, mceq_hdf_db, override_decay_db_name=None):
        #: MCEq HDF5Backend reference
        self.mceq_db = mceq_hdf_db
        #: (list) List of particles in the decay matrices
        self.parent_list = []
        self._default_decay_dset = override_decay_db_name

        if self._default_decay_dset is None:
            if config.muon_helicity_dependence:
                self._default_decay_dset = "polarized"
            else:
                self._default_decay_dset = "unpolarized"

    def load(self, parent_list=None, decay_dset=None):
        # Load tables and index from file
        if decay_dset is None:
            decay_dset = self._default_decay_dset

        index = self.mceq_db.decay_db(decay_dset)

        self.parents = index["parents"]
        self.particles = index["particles"]
        self.relations = index["relations"]
        self.index_d = index["index_d"]
        self.description = index["description"]
        # Advanced options
        regenerate_index = False
        if parent_list:
            # Take only the parents provided by the list
            # Add the decay products, which can become new parents
            def _follow_decay_chain(p, plist):
                if p in self.relations:
                    plist.append(p)
                    for d in self.relations[p]:
                        _follow_decay_chain(d, plist)
                else:
                    return plist

            plist = []
            for p in parent_list:
                _follow_decay_chain(p, plist)

            self.parents = sorted(list(set(plist)))
            regenerate_index = True

        if regenerate_index:
            self.particles = []
            for p in list(self.relations):
                if p not in self.parents:
                    _ = self.relations.pop(p, None)
                    continue
                self.particles.append(p)
                self.particles += self.relations[p]
            self.particles = sorted(list(set(self.particles)))

    def __getitem__(self, key):
        return self.get_matrix(*key)

    def __contains__(self, key):
        """Defines the `in` operator to look for particles"""
        return key in self.parents

    def children(self, parent_pdg):
        if parent_pdg not in self.relations:
            raise Exception(f"Parent {parent_pdg} not in decay database.")

        return self.relations[parent_pdg]

    def get_matrix(self, parent, child):
        """Returns a ``DIM x DIM`` decay matrix.

        Args:
          parent (int): PDG ID of parent particle
          child (int): PDG ID of final state child particle
        Returns:
          numpy.array: decay matrix
        """
        info(20, "entering with", parent, child)
        if child not in self.relations[parent]:
            raise Exception(f"trying to get empty matrix {parent} -> {child}")

        return self.index_d[(parent, child)]


class InteractionCrossSections:
    """Class for managing the dictionary of hadron-air cross-sections.

    Args:
      mceq_hdf_db (object): instance of :class:`MCEq.data.HDF5Backend`
      interaction_model (str): name of the interaction model
    """

    #: unit - :math:`\text{GeV} \cdot \text{fm}`
    GeVfm = 0.19732696312541853
    #: unit - :math:`\text{GeV} \cdot \text{cm}`
    GeVcm = GeVfm * 1e-13
    #: unit - :math:`\text{GeV}^2 \cdot \text{mbarn}`
    GeV2mbarn = 10.0 * GeVfm**2
    #: unit conversion - :math:`\text{mbarn} \to \text{cm}^2`
    mbarn2cm2 = GeVcm**2 / GeV2mbarn

    def __init__(self, mceq_hdf_db, interaction_model="DPMJETIII193"):
        #: MCEq HDF5Backend reference
        self.mceq_db = mceq_hdf_db
        #: reference to energy grid
        self.energy_grid = mceq_hdf_db.energy_grid
        #: List of active parents
        self.parents = None
        #: Dictionary containing the distribuiton matrices
        self.index_d = None
        #: (str) Interaction Model name
        self.iam = normalize_hadronic_model_name(interaction_model)

    def __getitem__(self, parent):
        """Return the cross section in :math:`\\text{cm}^2` as a dictionary
        lookup."""
        return self.get_cs(parent)

    def __contains__(self, key):
        """Defines the `in` operator to look for particles"""
        return key in self.parents

    def load(self, interaction_model):
        #: (str) Interaction Model name
        self.iam = normalize_hadronic_model_name(interaction_model)
        # Load tables and index from file
        index = self.mceq_db.cs_db(self.iam)

        self.parents = index["parents"]
        self.index_d = index["index_d"]

    def get_cs(self, parent, mbarn=False):
        """Returns inelastic ``parent``-air cross-section
        :math:`\\sigma_{inel}^{proj-Air}(E)` as vector spanned over
        the energy grid.

        Args:
          parent (int): PDG ID of parent particle
          mbarn (bool,optional): if ``True``, the units of the cross-section
                                 will be :math:`mbarn`,
                                 else :math:`\\text{cm}^2`

        Returns:
          numpy.array: cross-section in :math:`mbarn` or :math:`\\text{cm}^2`
        """

        message_templ = "replacing {0} with {1} cross section"
        if isinstance(parent, tuple):
            parent = parent[0]
        if parent in list(self.index_d):
            cs = self.index_d[parent]
        elif abs(parent) in list(self.index_d):
            cs = self.index_d[abs(parent)]
        elif 100 < abs(parent) < 300 and abs(parent) != 130:
            cs = self.index_d[211]
        elif 300 < abs(parent) < 1000 or abs(parent) in [130, 10313, 10323]:
            info(15, message_templ.format(parent, "K+-"))
            cs = self.index_d[321]
        elif abs(parent) > 1000 and abs(parent) < 5000:
            info(15, message_templ.format(parent, "nucleon"))
            cs = self.index_d[2212]
        elif 5 < abs(parent) < 23:
            info(15, "returning 0 cross-section for lepton", parent)
            return np.zeros(self.energy_grid.d)
        else:
            info(1, "Strange case for parent, using zero cross section.")
            cs = 0.0

        if not mbarn:
            return self.mbarn2cm2 * cs
        else:
            return cs


class ContinuousLosses:
    """Class for managing the dictionary of hadron-air cross-sections.

    Args:
      mceq_hdf_db (object): instance of :class:`MCEq.data.HDF5Backend`
      material (str): name of the material (not fully implemented)
    """

    def __init__(self, mceq_hdf_db):
        #: MCEq HDF5Backend reference
        self.mceq_db = mceq_hdf_db
        #: reference to energy grid
        self.energy_grid = mceq_hdf_db.energy_grid
        #: List of active parents
        self.parents = None
        #: Dictionary containing the distribuiton matrices
        self.index_d = None
        # Load defaults
        self.load_db()

    def __getitem__(self, parent):
        """Return the cross section in :math:`\\text{cm}^2` as
        a dictionary lookup."""
        return self.index_d[parent]

    def __contains__(self, key):
        """Defines the `in` operator to look for particles"""
        return key in self.parents

    def load_db(self):
        from scipy.interpolate import InterpolatedUnivariateSpline

        # Load tables and index from file
        index = self.mceq_db.continuous_loss_db()
        self.parents = index["parents"]
        self.index_d = index["index_d"]
        if "generic" not in index and config.generic_losses_all_charged:
            raise Exception("New data file needed to support generic losses.")
        self.generic_spl = InterpolatedUnivariateSpline(
            np.log(index["generic"][0]), np.log(index["generic"][1]), k=1
        )
